# -*- coding: utf-8 -*-

import datetime

from django.conf import settings
from django.contrib import admin
from django.core import exceptions
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.conf.urls.defaults import patterns,url
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

import django.db as db
import django.http as http
import django.forms as forms
import django.contrib.auth.models as auth_models
import django.contrib.auth.admin as auth_admin
from django.views.generic import TemplateView

import basecontacts_gabnor.references.models as ref_models
import basecontacts_gabnor.contacts.models as contacts_models
import basecontacts_gabnor.contacts.admin_forms as admin_forms
import basecontacts_gabnor.contacts.forms as contacts_forms
import basecontacts_gabnor.contacts.admin_views as contacts_views
import basecontacts_gabnor.contacts.utils as contacts_utils

from basecontacts_gabnor import logger


### --------------------------------------------------------------------------
# Adminsite
#


class SiteGestionContacts(admin.sites.AdminSite):
    """
    Personnalisation de AdminSite pour ajouter nos vues.
    """
    def get_urls(self):
        """
        """
        urls = super(SiteGestionContacts, self).get_urls()
        my_urls = patterns('',
            #(r'^test/$', self.my_view),

            url(r'^contacts/panier/(?P<ct_id>\d+)/afficher/$',
                self.admin_view(contacts_views.AffichePanier.as_view()),
                {}, name='affiche_panier'),

            url(r'^contacts/panier/(?P<ct_id>\d+)/vider/$',
                self.admin_view(contacts_views.VidePanier.as_view()),
                {}, name='vide_panier'),

            url(r'^contacts/panier/(?P<ct_id>\d+)/action/$',
                self.admin_view(contacts_views.ActionPanier.as_view()),
                {}, name='action_panier'),

            url(r'^contacts/structures/etiquettes/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenEtiquettes.as_view()),
                {}, name='gen_etiquettes'),

            url(r'^contacts/mailto/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenMailto.as_view()),
                {}, name='gen_mailto'),

            url(r'^contacts/exportcsv/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenCSVComplet.as_view()),
                {}, name='export_csv_complet'),

            url(r'^contacts/exportcsv/structures/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenCSVCompletStructures.as_view()),
                {}, name='export_csv_complet_structures'),

            url(r'^contacts/exportcsv/personnes/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenCSVCompletPersonnes.as_view()),
                {}, name='export_csv_complet_personnes'),

            url(r'^contacts/exportpublipostage/(?P<ct_id>\d+)/$',
                self.admin_view(contacts_views.GenCSVPublipostage.as_view()),
                {}, name='export_csv_publipostage'),

            url(r'^contacts/agriculteur/(?P<agriculteur_id>\d+)/transformer/$',
                self.admin_view(contacts_views.TransformeAgriculteur.as_view()),
                {}, name='transforme_agriculteur'),

            url(r'^contacts/agriculteurbio/(?P<agriculteurbio_id>\d+)/donneesbio/add/$',
                self.admin_view(contacts_views.CreeDonneesBio.as_view()),
                {}, name='cree_donneesbio'),

            url(r'^aide/$',
                self.admin_view(TemplateView.as_view(template_name="aide.html")),
                {}, name='aide'),

        )
        return my_urls + urls

    def my_view(self, request):
        """Vue de test."""
        return http.HttpResponse("Hello, world.")

site = SiteGestionContacts()




### --------------------------------------------------------------------------
# Admin des utilisateurs
#

class AdminUser(auth_admin.UserAdmin):
    """
    On redéfinit l'admin pour les utilisateurs pour éviter l'édition de champs
    inutiles. On hérite de auth_admin.UserAdmin qui définit les urls d'accès
    aux formulaire de changement de mot de passe, etc.
    """
    ordering = ('username',)
    list_display = ('username', 'first_name', 'last_name', 'is_active')
    list_display_links = ('username',)
    list_filter = ('is_active',)

    fieldsets = (
        (None, {'fields': ('username', 'password')}),
        (_('Personal info'), {'fields': ('first_name', 'last_name', 'email')}),
        (_('Permissions'), {'fields': ('is_active', )}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined')}),
    )
    readonly_fields = ('last_login', 'date_joined')

    def save_model(self, request, obj, form, change):
        """
        Redéfinition de save_model pour forcer tous les utilisateurs à faire
        partie du staff (pour accéder à la partie admin) et superutilisateur
        (pour pouvoir tout modifier) puisqu'on n'utilise pas de droit.
        """
        # on force aussi le username en minuscules pour éviter les ambiguités
        obj.username = obj.username.lower()
        obj.is_superuser = True
        obj.is_staff = True
        obj.save()




### ---------------------------------------------------------------------------
# Classes abstraites de base (éléments communs à d'autres classes)
#


class MixinAdminAvecPanier:
    """
    Class abstraite de base pour les pages d'admin permettant l'utilisation
    d'un panier.
    .
    La classe n'hérite pas de ModelAdmin pour pouvoir être utilisée
    conjointement avec MixinAdminFiltrable (et éviter un double héritage).
    .
    Procédure pour faire bénéficier un modèle d'un panier pour export CSV :
    * dans models.py, définir les attributs de classe du modèle suivants:
      - « classe_form_actions_panier » donnant la classe du form de choix
        d'action à utiliser (héritière de contacts_forms.BaseFormActionPanier)
        à choisir selon les actions qu'on veut permettre sur le panier
      - les attributs éventuellement nécessaires aux actions permises (en
        particulier « attrs_export » pour définir les attributs à inclure dans
        les exports CSV)
    * dans admin.py, faire hériter la classe d'admin de MixinAdminAvecPanier
    * créer un répertoire au nom du modèle (en minuscules) dans 
      « templates/admin/contacts », où on placera un template, même vide,
      héritant du template "admin/contacts/base_change_list.html".
    """
    # Actions sur la liste
    actions = [ 'ajouter_au_panier' ]
    
    def changelist_view(self, request, extra_context=None):
        """
        Surcharge de la vue de liste des objets pour ajouter le formulaire à
        afficher et le nombre de contacts dans le panier.
        """
        ct = ContentType.objects.get_for_model(self.model)
        mc = ct.model_class()
        # Préparation du contexte
        url_affiche_panier = reverse('admin:affiche_panier', args=[ct.pk])
        url_vide_panier = "%s?back=%s" % \
                    (reverse('admin:vide_panier', args=[ct.pk]),
                     reverse('admin:%s_%s_changelist' \
                             % (self.model._meta.app_label,
                                self.model.__name__.lower())))
        url_action_panier = reverse('admin:action_panier', args=[ct.pk])
        try:
            # Ne pas instancier le form ici car on pourrait masquer
            # une exception AttributeError dans son constructeur.
            form_actions_class = mc.classe_form_actions_panier
        except AttributeError:
            raise contacts_models.AppImproperlyConfigured(
                (u"Le modèle %s ne possède pas" % mc.__name__
                 + u" d'attribut classe_form_actions_panier"))
        form_actions = form_actions_class()

        if extra_context is None:
            extra_context = {}
        extra_context.update({
            'nom_panier': mc.basket_name,
            'cardinal_panier': self._compte_panier(request),
            'url_affiche_panier': url_affiche_panier,
            'url_vide_panier': url_vide_panier,
            'url_action_panier': url_action_panier,
            'form_action': form_actions,
        })
        return admin.ModelAdmin.changelist_view(self, request,
                                                extra_context=extra_context)
    
    def ajouter_au_panier(self, request, queryset):
        """
        Action ajoutant les contacts sélectionnés au panier.
        """
        ct = ContentType.objects.get_for_model(self.model)
        id_panier = self.model.basket_name
        if not request.session.has_key(id_panier):
            request.session[id_panier] = []
        for obj in queryset:
            request.session[id_panier].append((ct.id, obj.pk))
        request.session.modified = True

    ajouter_au_panier.short_description = \
            u"Ajouter les éléments sélectionnés au panier"

    def _compte_panier(self, request):
        """
        Retourne le nombre de contacts dans le panier courant, selon le modèle
        en cours.
        """
        id_panier = self.model.basket_name
        if not request.session.has_key(id_panier):
            return 0
        return len(request.session[id_panier])
        


class MixinAdminFiltrable(MixinAdminAvecPanier):
    """
    Classe abstraite de base pour les pages d'admin dont la liste est filtrable
    par nos filtres spéciaux.
    .
    Note: les filtres spéciaux sont conservés en session pour pouvoir être
    utilisés et filtrer le queryset même lorsqu'un autre form de la page est
    validé.
    C'est le cas lorsque l'utilisateur sélectionne les objets de toutes les
    pages de la liste via le lien « sélectionner tous les xxx agriculteurs »,
    puisque dans ce cas, django ne limite pas la sélection aux ids renvoyés par
    les cases cochées, et nos filtres spéciaux seraient ignorés parce que les
    critères de filtrage n'étaient pas visibles à notre fonction de filtrage au
    déclenchement de l'action puisque ces données font partie d'un form HTML
    différent (le formulaire « form_filtre » ne retrouvait aucune donnée le
    concernant dans request.POST, et donc ne filtrait rien).
    Les filtres spéciaux sont donc actifs tant qu'ils ne sont pas redéfinis ou
    supprimés via le lien « Supprimer tous les filtres » qui passe un argument
    GET spécial déclenchant la suppression des données.
    """
    # Form de filtrage à utiliser
    classe_form_filtre = None
    
    def changelist_view(self, request, extra_context=None):
        """
        Surcharge de la vue de liste des objets pour ajouter le formulaire à
        afficher et le nombre de contacts dans le panier.
        .
        Note: self.form_filtre n'est défini que dans le cas de l'affichage de la vue
        changelist, mais l'objet est réutilisé par le processus, donc il peut
        être présent même si changelist_view n'est pas appelée :/.
        """
        if extra_context is None:
            extra_context = {}
        if self.classe_form_filtre:
            if request.method == "POST" \
                and request.POST.get('filtres-speciaux') is not None:
                self.form_filtre = self.classe_form_filtre(request.POST)
            else:
                request.session.setdefault('filtres', {})
                filtres_data = request.session['filtres']\
                                    .get(self.model.__name__.lower(), {})
                self.form_filtre = self.classe_form_filtre(initial=filtres_data)
            extra_context["form_filtre"] = self.form_filtre
        return MixinAdminAvecPanier.changelist_view(self, request,
                                                    extra_context)

    def get_object(self, request, object_id):
        """
        On redéfinit ModelAdmin.get_object() pour ne pas utiliser les filtres
        dans queryset().
        """
        queryset = self.queryset(request, False)
        model = queryset.model
        try:
            object_id = model._meta.pk.to_python(object_id)
            return queryset.get(pk=object_id)
        except (model.DoesNotExist, ValidationError):
            return None

    def queryset(self, request, please_filter=True):
        """
        Surcharge de la définition du queryset pour appliquer les filtres
        choisis via le formulaire ajouté et enregistrés en session.
        .
        Note: queryset() est aussi appelé par get_object(). Pour éviter
        d'interdire la lecture d'un objet qui serait exclu par le filtre
        conservé en session, on ne filtre pas lorsque please_filter vaut False
        (c'est le cas quand il est appelé par get_object()).
        """
        if not hasattr(self, 'form_filtre') or not please_filter:
            # On n'affiche pas une liste d'objets, donc pas de filtre
            return admin.ModelAdmin.queryset(self, request)
        request.session.setdefault('filtres', {})
        filtres_data = request.session['filtres']\
                            .get(self.model.__name__.lower(), {})
        if request.method == "POST" \
                and request.POST.get('filtres-speciaux') is not None:
            # L'utilisateur vient de changer les filtres spéciaux
            if self.form_filtre.is_valid():
                filtres_data = self.form_filtre.cleaned_data
                request.session['filtres'][self.model.__name__.lower()] \
                                                            = filtres_data
                request.session.modified = True
            else:
                logger.warning("Filtrage de listes d'objets: formulaire"
                               " invalide (aucun filtrage effectué).")
        elif request.method == "GET" \
                and request.GET.get('supprfiltres', 0):
            # Si l'argument GET « supprfiltres » est présent et != 0,
            # on supprime les filtres mémorisés
            filtres_data = {}
            try:
                del request.session['filtres'][self.model.__name__.lower()]
            except KeyError:
                # évite les problèles lorsqu'aucun filtre n'est défini
                pass
            else:
                request.session.modified = True
        # Si le filtrage est actif, on le signale dans le form pour le
        # faire remonter dans le template par effet de bord (la méthode
        # queryset() est appelée trop tard pour modifier le contexte).
        if filtres_data:
            self.form_filtre.filtrant = True
        else:
            self.form_filtre.filtrant = False
        filtres_qs = self.form_filtre.construit_filtres_qs(filtres_data)
        return admin.ModelAdmin.queryset(self, request)\
                    .filter(*filtres_qs)\
                    .distinct()



class BaseAdminPersonne:
    """
    Classe de base pour l'administration des personnes, que ce soit en inline
    ou en propre.
    .
    La classe n'hérite pas directement de ModelAdmin pour pouvoir être utilisée
    avec des ModelAdmin que des Inline.
    """
    form = admin_forms.PersonneAdminForm
    readonly_fields = ('date_creation', 'date_derniere_modification',
                       'utilisateur_derniere_modification')
    fieldsets = (
            ('Désignation', {
                'fields': ('nom', 'prenom', 'structure', 'fonction',
                           'reseaux_engagements'),
                }),
            ('Coordonnées', {
                'fields': ('adresse', 
                           'commune', ('code_postal', 'ville'), 
                           'courrier_structure',
                           'tel_fixe', 'tel_mobile', 'fax', 'courriel',
                           'pas_de_courriel'),
                }),
            ('Enregistrement', {
                'fields': ('date_creation',
                           'date_derniere_modification',
                           'utilisateur_derniere_modification',),
                'classes': ('collapse',),
                }),
            )
   


# InlinePersonne doit être définie avant BaseAdminStructure
class InlinePersonne(BaseAdminPersonne, admin.StackedInline):
    """
    Classe de personnalisation de l'admin pour le modèle Personne en inline.

    Note: on ne peut pas hériter de MixinAdminFiltrable à ce niveau à cause
    car le fonctionnement n'est pas le même.
    """
    model = contacts_models.Personne
    extra = 0



class BaseAdminStructure(MixinAdminFiltrable, admin.ModelAdmin):
    """
    Classe de base pour la personnalisation de l'admin des structures et
    dérivées. À dériver pour les différentes structures implémentées.
    """
    class Media:
        css = {
            "all": ("css/contacts.css",),
        }
        js = ['js/list_filter_collapse.js']

    form = admin_forms.StructureAdminForm
    date_hierarchie = 'date_creation'
    list_display = ('id', 'nom_affichage', 'statut',
                    'categorie', 'commune', 'type' )
    list_display_links = ('nom_affichage',)
    list_filter = ('statut',
                   'categorie',
                   'commune__zone',
                   'commune__petite_region',
                   'commune__arrondissement',
                   )
    search_fields = ('designation', 'personne__nom', 'commune__nom',)
    readonly_fields = ('date_creation', 'date_derniere_modification',
                       'utilisateur_derniere_modification')
    fieldsets = (
            ('Désignation', {
                'fields': ('designation', 'siret', 'statut', 'categorie',
                           'commentaire', 'etiquettes'),
                }),
            ('Coordonnées', {
                'fields': ('adresse', 
                           'commune', ('code_postal', 'ville'),
                           'tel_fixe', 'tel_mobile', 'fax', 'courriel',
                           'site_web'), 
                'classes': ('collapse',),
                }),
            ('Contact', {
                'fields': ('origine_contact', 
                           ('annee_contact', 'mois_contact'),
                           ('date_creation', 'date_derniere_modification',
                           'utilisateur_derniere_modification'), ),
                'classes': ('collapse',),
                }),
            )
    filter_horizontal = ('etiquettes',)
    inlines = [ InlinePersonne, ]
    classe_form_filtre = contacts_forms.FormFiltreStructure

    def nom_affichage(self, obj):
        """
        """
        return str(obj)
    nom_affichage.short_description = "Structure ou personne"

    def save_model(self, request, obj, form, change):
        """
        Redéfinition de save_model pour assigner automatiquement le nom de
        l'utilisateur ayant effectué la modification.
        """
        obj.utilisateur_derniere_modification = request.user
        obj.save()

    def save_formset(self, request, form, formset, change):
        """
        Assignation de l'utilisateur ayant effectué la dernière modification
        des objets Personne en Inline (pour lesquel il n'y a pas de méthode
        save_model).
        """
        instances = formset.save(commit=False)
        for instance in instances:
            instance.utilisateur_derniere_modification = request.user
            instance.save()
        formset.save_m2m() 



### ---------------------------------------------------------------------------
# Admin de Personne
#

class AdminPersonne(BaseAdminPersonne, MixinAdminFiltrable, admin.ModelAdmin):
    """
    Classe de personnalisation de l'admin pour le modèle Personne.
    """
    list_display = ('nom', 'prenom', 'structure',)
    list_display_links = ('nom','prenom', 'structure')
    #list_filter = (,)
    search_fields = ('nom', 'prenom',)

    def save_model(self, request, obj, form, change):
        """
        Redéfinition de save_model pour assigner automatiquement le nom de
        l'utilisateur ayant effectué la modification.
        """
        obj.utilisateur_derniere_modification = request.user
        obj.save()




### ---------------------------------------------------------------------------
# Admin de StructuresGenerique
#

class AdminStructure(BaseAdminStructure):
    """
    Classe de personnalisation de l'admin pour le modèle Structure.
    """

    actions = BaseAdminStructure.actions + [ 'transformer_en_agri' ]

    def transformer_en_agri(self, request, queryset):
        """
        Action permettant de transformer les structure cochées en
        agriculteurs.
        """
        msg_unique = u"La structure sélectionnée a été transformée"\
                    u" en agriculteurs."
        msg_plural = u"Les structures sélectionnées ont été transformées"\
                    u" en agriculteurs."
        resultat = 0
        for structgen in queryset:
            resultat += contacts_utils.transforme_structure(structgen)
        db.transaction.commit_unless_managed()
        if len(queryset) == 1:
            if resultat == 1:
                msg = msg_unique
            else:
                msg = u"La structure sélectionnée est déjà un agriculteur !"
        else:
            msg = msg_plural
            if resultat != len(queryset):
                msg += u" (note: au moins une structure sélectionnée"\
                        u" était déjà un agriculteur)"
        messages.success(request, msg)

    transformer_en_agri.short_description = \
            u"Transformer les structures sélectionnées en agriculteur"


### -------------------------------------------------------------------------
# Admin de Agriculteur et dépendances
#


class InlineProjetBio(admin.StackedInline):
    """
    Inline pour affichage des projets bio dans le formulaire d'un Agriculteur.
    """
    model = contacts_models.ProjetBio
    extra = 0
    
    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        On exclut l'utilisateur « admin » de la liste des techniciens.
        """
        if db_field.name == "technicien":
            kwargs["queryset"] = auth_models.User.objects\
                                                .exclude(username='admin')
        return super(InlineProjetBio, self).formfield_for_foreignkey(\
                                                db_field, request, **kwargs)

#    def delete_model(self, request, obj):
#        """
#        """
#        print "DELETING %s" % obj
        



class InlineAgriRendezVousProjetBio(admin.StackedInline):
    """
    Inline pour affichage des rendez-vous de projets bio dans le formulaire
    d'un Agriculteur (à défaut de pouvoir les afficher au sein des inline des
    projet bio, puisque Django ne supporte pas les inline imbriqués).
    .
    Une clef étrangère agriculteur a été ajoutée pour ça dans le modèle (en
    plus de la clef vers le projet bio) pour pouvoir inclure ces rendez-vous
    sur la page de l'agriculteur.
    À noter: ce hack oblige à supprimer les rendez-vous de projet bio
    avant de supprimer le projet bio qui s'y rapporte.
    """
    model = contacts_models.RendezVousProjetBio
    verbose_name = "rendez-vous de suivi de projet bio"
    verbose_name_plural = "rendez-vous de suivi de projet bio"
    extra = 0
    fieldsets = (
            (None, {
                'fields': (('projet_bio', 'date',),
                           ('type', 'technicien',),
                           'commentaire'),
                }),
            )

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Permet de modifier le champ de formulaire utilisé pour une clef
        étrangère. En l'occurence, on cherche à limiter la liste des projets
        bio à ceux de l'agriculteur concerné.
        Il s'agit d'un hack affreux parce qu'à ce niveau, on a pas accès à
        l'agriculteur qu'on est en train d'éditer. Mais... on peut en récupérer
        son id directement depuis l'URL (je l'avais dit que c'était affreux).
        C'est fragile en cas de modification de l'appli, mais ça marche et ça
        évite aux utilisateurs d'avoir à sélectionner dans une liste longue
        comme le bras (ou à passer uniquement par la page d'édition d'un projet
        bio, où il n'y a pas besoin de faire le choix).
        Ça marche aussi dans le cas des pages de modification d'agriculteurbio
        car l'id de agriculteurbio, agriculteur (et même structure) sont le même.
        """
        formfield = super(InlineAgriRendezVousProjetBio, self)\
                        .formfield_for_foreignkey(db_field, request, **kwargs)
        if db_field.name == "projet_bio":
            path_info = request.META["PATH_INFO"]
            try:
                id_agri = int(path_info.split('/')[-2])
            except ValueError:
                # Il ne s'agit pas d'un entier, mais de "add", etc.
                pass
            else:
                formfield.queryset = formfield.queryset\
                                    .filter(agriculteur__id=id_agri)
        return formfield



class AdminAgriculteur(BaseAdminStructure):
    """
    Classe de personnalisation de l'admin pour le modèle Agriculteur.
    """
    form = admin_forms.AgriculteurAdminForm
    date_hierarchie = 'date_creation'
    list_display = ('id', 'nom_affichage', 'statut', 'commune', 'type')
    list_display_links = ('nom_affichage',)
    list_filter = ('statut',
                   'type_activite',
                   'commune__zone',
                   'commune__petite_region',
                   'commune__arrondissement',
                   )
    search_fields = ('designation', 'personne__nom', 'commune__nom',)
    readonly_fields = ('date_creation', 'date_derniere_modification',
                       'utilisateur_derniere_modification')
    filter_horizontal = BaseAdminStructure.filter_horizontal + \
                            ('qualificatifs_complementaires',)
    fieldsets = (
            ('Désignation', {
                'fields': ('designation', 'siret', 'statut', 'commentaire',
                           'visibilite', 'etiquettes'),
                }),
            ('Coordonnées', {
                'fields': \
                    dict(BaseAdminStructure.fieldsets)['Coordonnées']['fields'],
                'classes': ('collapse',),
                }),
            ('Informations sur l\'exploitation', {
                'fields': ( 'type_activite', 'qualificatifs_complementaires',
                            ('sau','sau_bio'), ('sau_eau', 'sau_eau_bio'),
                            ('quota_laitier', 'laiterie',
                             'nb_vaches_allaitantes',), ),
                'classes': ('collapse',),
                }),
            ('Contact', dict(BaseAdminStructure.fieldsets)['Contact']),
            )
    inlines = ( InlinePersonne,
                InlineProjetBio,
                InlineAgriRendezVousProjetBio, )
    classe_form_filtre = contacts_forms.FormFiltreAgriculteur
    actions = BaseAdminStructure.actions + [ 'transformer_en_bio' ]

    def transformer_en_bio(self, request, queryset):
        """
        Action permettant de transformer les agriculteurs cochés en
        agriculteurs biologiques.
        """
        msg_unique = u"L'agriculteur sélectionné a été transformé"\
                    u" en agriculteurs bio."
        msg_plural = u"Les agriculteurs sélectionnés ont été transformés"\
                    u" en agriculteurs bio."
        resultat = 0
        for agrigen in queryset:
            resultat += contacts_utils.transforme_agriculteur(agrigen)
        db.transaction.commit_unless_managed()
        if len(queryset) == 1:
            if resultat == 1:
                msg = msg_unique
            else:
                msg = u"L'agriculteur sélectionné est déjà bio !"
        else:
            msg = msg_plural
            if resultat != len(queryset):
                msg += u" (note: au moins un agriculteur sélectionné"\
                        u" était déjà bio)"
        messages.success(request, msg)

    transformer_en_bio.short_description = \
            u"Transformer les agriculteurs sélectionnés en agriculteur biologique"


### ---------------------------------------------------------------------------
# Admin de AgriculteurBio et dépendances
#


class InlineAdhesionGabnor(admin.TabularInline):
    """
    """
    model = contacts_models.AdhesionGabnor
    verbose_name = "adhésion GABnor"
    verbose_name_plural = "adhésions GABnor"
    extra = 0
    fieldsets = (
            (None, {
                'fields': (('date', 'montant'),),
                }),
            )


class InlineDonneesBio(admin.StackedInline):
    """
    """
    model = contacts_models.DonneesBio
    verbose_name = "données annuelles de l'observatoire"
    verbose_name_plural = "données annuelles de l'observatoire"
    extra = 0
    fieldsets = (
            )



class InlineRendezVousSuivi(admin.StackedInline):
    """
    """
    model = contacts_models.RendezVousSuivi
    verbose_name = "rendez-vous"
    verbose_name_plural = "rendez-vous"
    extra = 0
    fieldsets = (
            (None, {
                'fields': (('technicien', 'date'), 'commentaire'),
                }),
            )



class AdminAgriculteurBio(BaseAdminStructure):
    """
    Classe de personnalisation de l'admin pour le modèle AgriculteurBio.
    """
    form = admin_forms.AgriculteurBioAdminForm
    date_hierarchie = 'date_creation'
    list_display = ('id', 'nom_affichage', 'statut', 'commune', 'get_date_premiere_conversion')
    list_display_links = ('nom_affichage',)
    list_filter = ('statut',
                   'type_activite',
                   'commune__zone',
                   'commune__petite_region',
                   'commune__arrondissement',
                   )
    filter_horizontal = BaseAdminStructure.filter_horizontal + \
                            ('qualificatifs_complementaires',)
    search_fields = ('designation', 'personne__nom', 'commune__nom',)
    readonly_fields = ('date_creation', 'date_derniere_modification',
                       'utilisateur_derniere_modification',
                       'premiere_adhesion',
                       'renouvellement_adhesion')
    fieldsets = (
            ('Désignation',
                dict(AdminAgriculteur.fieldsets)['Désignation']),
            ('Coordonnées',
                dict(AdminAgriculteur.fieldsets)['Coordonnées']),
            ('Informations sur l\'exploitation', 
                dict(AdminAgriculteur.fieldsets)\
                        ['Informations sur l\'exploitation']),
            ('Contact', dict(AdminAgriculteur.fieldsets)['Contact']),
            ('Infos adhésions Gabnor', {
                'fields': (
                           'date_envoi_courrier',
                           'premiere_adhesion',
                           'renouvellement_adhesion'),
                'classes': ('collapse',),
                }),
            )
    inlines = ( InlineAdhesionGabnor,
                InlinePersonne,
                InlineProjetBio,
                InlineAgriRendezVousProjetBio,
                InlineRendezVousSuivi,
                )
    classe_form_filtre = contacts_forms.FormFiltreAgriculteurBio

    def save_model(self, request, obj, form, change):
        """
        Avant de sauvegarder l'objet, met à jour le champ sau_agri de
        l'observatoire de l'année en cours, s'il existe.
        On ne fait rien si l'objet n'est pas encore enregistré.
        """
        if obj.id:
            try:
                dbio = obj.donneesbio_set.get(annee=datetime.date.today().year)
            except contacts_models.DonneesBio.DoesNotExist, e:
                pass
            else:
                dbio.sau_agri = obj.sau
                dbio.save()
        super(AdminAgriculteurBio, self).save_model(request, obj, form, change)



### ---------------------------------------------------------------------------
# Admin de ProjetBio et dépendances
#

class InlineRendezVousProjetBio(admin.StackedInline):
    """
    Inline pour affichage des rendez-vous de projets bio dans le formulaire
    des projets bio.
    """
    model = contacts_models.RendezVousProjetBio
    verbose_name = "rendez-vous de suivi de projet bio"
    verbose_name_plural = "rendez-vous de suivi de projet bio"
    extra = 0
    exclude = ('agriculteur',)
    fieldsets = (
            (None, {
                'fields': (('projet_bio', 'date',),
                           ('type', 'technicien',),
                           'commentaire'),
                }),
            )





class AdminProjetBio(MixinAdminAvecPanier, admin.ModelAdmin):
    """
    Classe d'admin pour afficher les projets bio de façon indépendante.
    """
#    list_display = ('id', 'agriculteur', '__unicode__', 'agriculteur_id')
    list_display = ('__unicode__', 'agri_id')
    list_display_links = ('__unicode__', )
    inlines = ( InlineRendezVousProjetBio, )
    search_fields = ('agriculteur__designation', 
                     'agriculteur__personne__nom',
                     'agriculteur__commune__nom')
    list_filter = ('technicien',)
    
    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "technicien":
            kwargs["queryset"] = auth_models.User.objects\
                                                .exclude(username='admin')
        return super(AdminProjetBio, self).formfield_for_foreignkey(\
                                                db_field, request, **kwargs)



### ---------------------------------------------------------------------------
# Admin de DonneesBio
#

class FormAssolement(forms.ModelForm):
    """
    Petite redéfinition du formulaire des assolements pour
    classer la liste de choix des cultures par catégories.
    """
    class Meta:
        model = contacts_models.Assolement

    def __init__(self, *args, **kwargs):
        """
        """
        super(FormAssolement, self).__init__(*args, **kwargs)
        self.fields['culture'].choices = self._get_cultures_as_choice()

    def _get_cultures_as_choice(self):
        """
        Retourne la liste des cultures organisée de façon à afficher les
        catégories.
        """
        choices = []
        for cat in ref_models.CATEGORIES_CULTURE:
            cultures = [
                ( c.id, c.designation.capitalize() )
                for c in ref_models.Culture.objects\
                        .filter(categorie=cat.lower()).order_by('indice_ordre') ]
            choices.append( (cat.capitalize(), cultures) )
        return choices


#class InlineAssolement(admin.TabularInline):
class InlineAssolement(admin.StackedInline):
    """
    Classe d'admin pour les données d'assolement de l'observatoire.
    """
    model = contacts_models.Assolement
    form = FormAssolement
    verbose_name = "Assolement"
    verbose_name_plural = "Assolements"
    extra = 0
    fieldsets = (
            (None, {
                'fields': (
                   ('culture','detail_culture'),
                   ('surface_bio',
                    'surface_c1','surface_c2','surface_c3',
                    'surface_conv'),
                   ('tonnage', 'prix', 'client', 'type_circuit'),
                   )
                }),
            )

class AdminDonneesBio(MixinAdminAvecPanier, admin.ModelAdmin):
    """
    Classe d'admin pour les données de l'observatoire (également éditées
    dans les formulaires des agriculteurs bio).
    """
    inlines = ( InlineAssolement, )
    search_fields = ('agriculteur_bio__designation', 
                     'agriculteur_bio__personne__nom',
                     'agriculteur_bio__commune__nom')
    list_filter = ('annee', )
    list_display = ('__unicode__', 'agri_id')
    # Ne fonctionne pas correctement -> à voir
    formfield_overrides = {
            db.models.BooleanField: { 'widget': forms.CheckboxInput },
        }
    filter_horizontal = ()
    readonly_fields = ('sau_agri', 'total_surfaces_assolement')
    fieldsets = (
            (None, {
                'fields': (
                   ('agriculteur_bio', # Problème: liste énorme :/
                   'annee'),
                   'organisme_certification',
                   ('mixite', 'biodynamie'),
                   )
                }),
            (u"Aides", {
                'fields': (
                   ('aide_conversion_ab', 'date_aide_conversion_ab'),
                   ('aide_maintien_ab', 'aide_credit_impot',
                    'aide_certification_bio'),
                   ('aide_investissement', 'aide_installation',
                    'aide_mae_t'),
                   'aide_autres',
                   ('aide_accompagnement', 'aide_accompagnement_precisions'),
                   ),
                'classes': ('wide', 'collapse',),
                }),
            (u"Main d'œuvre", {
                'fields': (
                   ('etp_associes', 'etp_familiaux'),
                   ('etp_salaries_permanents', 'etp_saisonniers'),
                   ('etp_encadrants', 'personnes_insertion'),
                   'remarques',
                   ),
                'classes': ('wide', 'collapse',),
                }),
            (u"Commercialisation", {
                'fields': (
                   ('vente_a_la_ferme', 'vente_ferme_detail'),
                   ('produits_dispo_marches', 'vente_marches_detail'),
                   ('produits_dispo_biocabas', 'biocabas_detail'),
                   ('produits_dispo_magasins', 'magasins_detail'),
                   ('produits_dispo_restaurateurs', 'restaurateurs_detail'),
                   ('restau_collective', 'restau_collective_detail'),
                   ('commerc_cl', 'commerc_cl_detail'),
                   ('statut_commercialisation', 'statut_choisi'),
                   ),
                'classes': ('wide', 'collapse',),
                }),
            (u'Accueil', {
                'fields': ('types_services', 'detail_activites', ),
                'classes': ('wide', 'collapse',),
                }),
            (u'Légumes', {
                'fields': (
                    'legumes',
                    'legumes_cultives_annee',
                    ('legumes_commercialisation_vd', 'legumes_vd_autres'),
                    ('legumes_commercialisation_cc', 'legumes_cc_autres'),
                    ('legumes_commercialisation_cl', 'legumes_cl_autres'),
                    'legumes_hors_vd',
                    'legumes_transformation',
                    'legumes_detail_transformation',
                    'legumes_systeme_forcage_endives',
                    ),
                'classes': ('wide', 'collapse',),
                }),
            (u"Cultures", {
                'fields': (
                    ('culture', 'culture_comm_vd',
                     'culture_comm_cc', 'culture_comm_cl'),
                    'culture_hors_vd',
                    ('culture_transformation','culture_detail_transformation'),
                    'culture_stockage_ferme',
                   ),
                'classes': ('wide', 'collapse',),
                }),
            (u"Vergers basse tige", {
                'fields': (
                    ('vergers_bt', 'vergers_bt_fruits'),
                    ('vergers_bt_comm_vd', 'vergers_bt_comm_cc',
                     'vergers_bt_comm_cl'),
                    'vergers_bt_hors_vd',
                    ('vergers_bt_transformation',
                     'vergers_bt_detail_transformation'),
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Vergers haute tige", {
                'fields': (
                    ('vergers_ht', 'vergers_ht_pommes',
                     'vergers_ht_poires', 'vergers_ht_prunes'),
                    ('vergers_ht_comm_vd', 'vergers_ht_comm_cc',
                     'vergers_ht_comm_cl'),
                    'vergers_ht_hors_vd',
                    ('vergers_ht_transformation',
                     'vergers_ht_detail_transformation'),
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Autres productions", {
                'fields': (
                    ('autres_petits_fruits', 'autres_petits_fruits_detail'),
                    ('autres_plants_maraichers', 'autres_plants_detail'),
                    'autres_nb_plants_an',
                    ('autres_ppam', 'autres_ppam_detail'),
                    ('autres_prod_veg', 'autres_prod_veg_detail'),
                    ('autres_comm_vd', 'autres_comm_cc', 'autres_comm_cl'),
                    'autres_hors_vd',
                    ('autres_transformation', 'autres_detail_transformation'),
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Lait de vaches", {
                'fields': (
                    # lait
                    ('lait_vaches', 'lait_nb_vaches', 'lait_vaches_conversion'),
                    ('lait_quota_laiterie', 'lait_quota_vente_directe'),
                    ('lait_prod_laiterie', 'lait_prod_vente_directe'),
                    ('lait_comm_vd', 'lait_comm_cc', 'lait_comm_cl'),
                    'lait_hors_vd',
                    'lait_laiterie',
                    ('lait_transformation_ferme', 'lait_transformation_presta'),
                    'lait_atelier_lait',
                    'lait_produits_transformes',
                    # viande issue de vaches laitières
                    'lait_reformes_cl',
                    'lait_genisses_cl',
                    'lait_taurillons_cl',
                    'lait_boeufs_cl',
                    'lait_caissette',
                    'lait_adultes_cc',
                    'lait_veaux_cc',
                    'lait_abattoirs_vl',
                    'lait_decoupe_presta',
                    'lait_atelier_decoupe',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Lait de chèvres", {
                'fields': (
                    ('lait_chevres', 'lait_nb_chevres', 'lait_chevres_conversion'),
                    'lait_chevres_prod',
                    ('lait_chevres_transformation_ferme',
                     'lait_chevres_transformation_presta'),
                    'lait_chevres_atelier_lait',
                    'lait_chevres_type_produits',
                    ('lait_chevres_comm_vd', 'lait_chevres_comm_cc',
                     'lait_chevres_comm_cl'),
                    'lait_chevres_hors_vd',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Viande bovine", {
                'fields': (
                    ('bovine', 'bovine_nb_va', 'bovine_nb_conversion'),
                    ('bovine_comm_vd', 'bovine_comm_cc', 'bovine_comm_cl'),
                    'bovine_hors_vd',
                    'bovine_reformes_cl',
                    'bovine_genisses_cl',
                    'bovine_taurillons_cl',
                    'bovine_boeufs_cl',
                    'bovine_caissette',
                    'bovine_adultes_cc',
                    'bovine_veaux_cc',
                    'bovine_abattoirs_va',
                    'bovine_decoupe_presta',
                    'bovine_atelier_va',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Porcs", {
                'fields': (
                    ('porcs', 'porcs_nb_truies', 'porcs_nb_conversion'),
                    ('porcs_comm_vd', 'porcs_comm_cc', 'porcs_comm_cl'),
                    'porcs_hors_vd',
                    'porcs_transformation_ferme',
                    'porcs_produits_transformes',
                    'porcs_abattoirs',
                    'porcs_decoupe_presta',
                    'porcs_atelier',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Moutons", {
                'fields': (
                    ('moutons', 'moutons_nb_brebis', 'moutons_nb_conversion'),
                    ('moutons_comm_vd', 'moutons_comm_cc', 'moutons_comm_cl'),
                    'moutons_hors_vd',
                    'moutons_transformation_ferme',
                    'moutons_abattoirs',
                    'moutons_decoupe_presta',
                    'moutons_atelier',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Petits animaux", {
                'fields': (
                    ('animaux_lapins', 'animaux_volailles'),
                    'animaux_volailles_chair',
                    'animaux_nb_conversion',
                    'animaux_nb_poulets',
                    'animaux_nb_lapins',
                    'animaux_nb_autres_volailles',
                    ('animaux_comm_vd', 'animaux_comm_cc', 'animaux_comm_cl'),
                    'animaux_hors_vd',
                    'animaux_atelier',
                    'animaux_decoupe_presta',
                    'animaux_lieu_abattage',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Autres productions animales", {
                'fields': (
                    'animaux_autres_prod',
                    'animaux_details_autres_prod',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Œufs", {
                'fields': (
                    'oeufs_production_bio',
                    'oeufs_nb_pondeuses', 'oeufs_nb_oeufs',
                    'oeufs_acheteur_cl',
                    ('oeufs_comm_vd', 'oeufs_comm_cc', 'oeufs_comm_cl'),
                    'oeufs_circuits_courts',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Miel", {
                'fields': (
                    'miel_production_bio',
                    'miel_nb_ruches',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Conventionnel", {
                'fields': (
                    ('conv_prod_animales', 'conv_detail_prod_animales'),
                    ('conv_prod_vegetales', 'conv_detail_prod_vegetales'),
                    ('conv_existe_prod_sans_valorisation',
                     'conv_prod_sans_valorisation'),
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Formation", {
                'fields': (
                    ('formation_annee', 'formation_annee_laquelle'),
                    'formation_annee_commentaire',
                    ('formation_avenir', 'formation_avenir_type'),
                    'formation_avenir_details',
                   ),
                'classes': ('wide', 'collapse'),
                }),
            (u"Surface d'assolement", {
                'fields': (
                    ('sau_agri', 'total_surfaces_assolement'),
                   ),
                'classes': ('wide',),
                }),
        )

    def save_model(self, request, obj, form, change):
        """
        Redéfinition de save_model pour mettre à jour automatiquement
        l'attribut sau_agri.
        """
        obj.sau_agri = obj.agriculteur_bio.sau
        obj.save()

    def save_formset(self, request, form, formset, change):
        """
        Recalcule le nouveau total des surfaces saisies dans les assolements
        après l'enregistrement des objets Assolement.
        Cette méthode n'est pas appelée si aucun assolement n'est modifiée.
        """
        # Mise à jour des objets assolements
        instances = formset.save(commit=True)
        # Calcul du nouveau total
        total = 0
        for a in form.instance.assolement_set.all():
            total += ((a.surface_bio or 0)
                      + (a.surface_c1 or 0)
                      + (a.surface_c2 or 0)
                      + (a.surface_c3 or 0)
                      + (a.surface_conv or 0))
        # Sauvegarde du nouveau total dans l'objet parent
        form.instance.total_surfaces_assolement = total
        form.instance.save()



### ---------------------------------------------------------------------------
# Admin de AdhesionGabnor
#

class AdminAdhesionGabnor(MixinAdminAvecPanier, admin.ModelAdmin):
    """
    Classe d'admin pour les adhésions Gabnor (également éditées
    dans les formulaires des agriculteurs bio).
    """
    date_hierarchy = 'date'
    list_filter = ('date',)
    search_fields = ('agriculteur_bio__designation', 
                     'agriculteur_bio__personne__nom',
                     'agriculteur_bio__commune__nom')
    

class AdminRendezVousProjetBio(MixinAdminAvecPanier, admin.ModelAdmin):
    search_fields = ('projet_bio__agriculteur__personne__nom',
                     'projet_bio__agriculteur__commune__nom',
                     'projet_bio__agriculteur__designation')
    list_filter = ('projet_bio__technicien',)
    list_display = ('projet_bio','agri_id', 'date')
    date_hierarchy = 'date'


### ---------------------------------------------------------------------------
# Enregistrement des modèles à administrer
#

def register_models(site):
    """
    Enregistrement des modèles dans un AdminSite passé en paramètre.
    """
    # Modèle User avec admin personnalisée
    site.register(auth_models.User, AdminUser)

    # Modèles principaux indispensables
    site.register(contacts_models.Structure, AdminStructure)
    site.register(contacts_models.Agriculteur, AdminAgriculteur)
    site.register(contacts_models.AgriculteurBio, AdminAgriculteurBio)
    site.register(contacts_models.Personne, AdminPersonne)

    # Facultatif: administration indépendante de données qu'on retrouve dans en
    # inline dans les autres pages.
    site.register(contacts_models.ProjetBio, AdminProjetBio)
    site.register(contacts_models.DonneesBio, AdminDonneesBio)
    site.register(contacts_models.AdhesionGabnor, AdminAdhesionGabnor)
    site.register(contacts_models.RendezVousProjetBio, AdminRendezVousProjetBio)
    

# fait dans urls.py
#register_models(site)


#
# Fin
### ---------------------------------------------------------------------------
