# -*- coding: utf-8 -*-

try: # On utilise cStringIO si possible
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import inspect

import django.http as http
import django.forms as forms
import django.db.models as models
import django.views.generic as generic_views

from django.conf import settings
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.contrib.contenttypes.models import ContentType

import basecontacts_gabnor.contacts.models as contacts_models
import basecontacts_gabnor.contacts.forms as contacts_forms
import basecontacts_gabnor.contacts.utils as contacts_utils
import basecontacts_gabnor.references.models as ref

from basecontacts_gabnor import logger
import re


### ---------------------------------------------------------------------------
# Vue d'affichage du panier
#

class AffichePanier(generic_views.list.ListView):
    """
    """
    template_name = "admin/contacts/affiche_panier.html"
    paginate_by = 50

    def get_queryset(self):
        """
        Retourne le queryset des objets à afficher (il apparaissent dans le
        template sous le nom « object_list »).
        """
        self.content_type = get_object_or_404(ContentType, pk=self.kwargs['ct_id'])
        self.model_class = self.content_type.model_class()
        selected = self.request.session.get(self.model_class.basket_name, [])
        object_list = []
        for ct_id, oid in selected:
            obj = get_object_or_404(ContentType, pk=ct_id)\
                               .get_object_for_this_type(pk=oid)
            obj.display_name = obj.__class__.__name__
            object_list.append(obj)
        return object_list

    def get_context_data(self, **kwargs):
        """
        Peuple le contexte du template.
        """
        context = super(AffichePanier, self).get_context_data(**kwargs)
        try:
            # Ne pas instancier le form ici car on pourrait masquer
            # une exception AttributeError dans son constructeur.
            form_actions_class = self.model_class.classe_form_actions_panier
        except AttributeError:
            # Si le modèle n'a pas de .classe_form_actions_panier,
            # c'est que l'appli est mal foutue :)
            raise contacts_models.AppImproperlyConfigured(
                ("Le modèle %s ne possède pas" % self.model_class.__name__
                 + " d'attribut classe_form_actions_panier"))
        url_changelist = reverse("admin:%s_%s_changelist" \
             % (self.content_type.app_label, self.content_type.model.lower()))
        url_vider_panier = reverse('admin:vide_panier',
                                   args=[ self.content_type.id ])\
                                    + "?back=%s" % url_changelist
        context.update({
            'url_action_form': reverse('admin:action_panier',
                                       args=[ self.content_type.id ]),
            'form': form_actions_class(),
            'nom_objet': { 
                'singulier': self.model_class._meta.verbose_name,
                'pluriel': self.model_class._meta.verbose_name_plural,
                },
            'nom_panier': self.model_class.basket_name,
            'url_vider_panier': url_vider_panier,
            'url_retour': url_changelist,
            })
        return context



### ---------------------------------------------------------------------------
# Vue de vidage du panier
#

class VidePanier(generic_views.base.RedirectView):
    """
    """

    def get(self, request, *args, **kwargs):
        """
        """
        ct = get_object_or_404(ContentType, pk=self.kwargs['ct_id'])
        request.session[ct.model_class().basket_name] = []
        # L'objet request.session n'a pas été modifié en lui-même, mais son
        # contenu l'a été. Il faut donc le préciser explicitement :
        request.session.modified = True
        messages.success(request, "Votre panier a été vidé")
        return generic_views.base.RedirectView.get(self, request,
                                                   *args, **kwargs)

    def get_redirect_url(self, **kwargs):
        """
        Retourne l'URL vers laquelle la vue va renvoyer l'utilisateur.
        On utilise l'argument CGI « back », ou on renvoit vers l'index s'il
        n'est pas présent.
        """
        return self.request.GET.get('back', reverse('admin:index'))



### ---------------------------------------------------------------------------
# Vue d'action sur les contacts du panier
#

class ActionPanier(generic_views.base.RedirectView):
    """
    Vue redirigeant l'utilisateur vers la vue effectuant l'action choisie.
    """

    def _get_classe_form_actions(self):
        """
        Détermine et retourne le formulaire à utiliser selon le modèle,
        dont la classe est fournie via le content type passé en URL.
        """
        self.content_type = get_object_or_404(ContentType,
                                              pk=self.kwargs['ct_id'])
        mc = self.content_type.model_class()
        try:
            return mc.classe_form_actions_panier
        except AttributeError:
            # Si le modèle n'a pas de .classe_form_actions_panier,
            # c'est que l'appli est mal foutue :)
            raise contacts_models.AppImproperlyConfigured(
                ("Le modèle %s ne possède pas" % self.content_type.model
                 + " d'attribut classe_form_actions_panier"))

    def post(self, request, *args, **kwargs):
        """
        Traite le résultat du formulaire et conserve l'action choisie dans
        self.action, qui sera utilisée par get_redirect_url().
        """
        form_class = self._get_classe_form_actions()
        form = form_class(request.POST)
        if not form.is_valid():
            # le formulaire ne devrait être invalide qu'en cas de bidouille
            logger.error("form action panier non valide")
            raise http.Http404(u"Les paramètres d'action sur"
                               u" la sélection sont incorrects")
        self.action = form.cleaned_data['action']
        return super(ActionPanier, self).get(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        """
        En cas d'appel par GET, on affiche une page 404.
        """
        raise http.Http404(u"Cette page peut être utilisée"
                           u" uniquement par un formulaire")

    def get_redirect_url(self, **kwargs):
        """
        Retourne l'URL vers laquelle la vue va renvoyer l'utilisateur.
        """
        return reverse("admin:"+self.action,
                       args=[ int(self.kwargs['ct_id']) ])



### ---------------------------------------------------------------------------
# Vues de génération des étiquettes
# 

class GenEtiquettes(generic_views.base.View):
    """
    Vue générant des étiquettes d'adresses des structures sélectionnées.
    """
    classe_vue_etiquettespdf = None

    def get(self, request, ct_id):
        """
        Retourne un document PDF où figure une étiquette pour chacun des objets
        sélectionnés.
        .
        Paramètres:
        ct_id: l'identifiant de l'object ContentType représentante le modèle des
               objets sélectionnés (on en déduit le panier à utiliser).
        Retourne:
        Le document PDF des étiquettes d'adresses.
        """
        ct = get_object_or_404(ContentType, pk=int(ct_id))
        model_class = ct.model_class()
        try:
            selection = request.session[model_class.basket_name]
        except KeyError:
            selection = []
        liste_contacts = []
        for ct_id, oid in selection:
            # On ajoute l'objet contact dans la liste de la sélection
            contact = ContentType.objects.get(pk=ct_id)\
                            .get_object_for_this_type(pk=oid)
            liste_contacts.append(contact)
            # On ajoute aussi les personnes associées qui doivent recevoir le
            # courrier (si le contact est déjà une personne, il n'y en a pas)
            for pers_associees in contact.liste_contacts_associes():
                if pers_associees.courrier_structure:
                    liste_contacts.append(pers_associees)
        if len(liste_contacts) == 0:
            raise http.Http404(u"Aucun contact n'a été trouvé !")
        # Génération et renvoi du document
        buffer = StringIO()
        pdfdoc = contacts_utils.EtiquettesPDF(buffer)
        pdfdoc.build(liste_contacts)
        response = http.HttpResponse(mimetype='application/pdf')
        response['Content-Disposition'] = 'attachment; filename=etiquettes.pdf'
        response.write(buffer.getvalue())
        buffer.close()
        return response



### ---------------------------------------------------------------------------
# Vue de préparation à l'envoi de courriel
#

class GenMailto(generic_views.base.TemplateView):
    """
    Vue générant un lien permettant d'envoyer un message à toutes les
    structures ou personnes sélectionnées.
    """
    template_name = "admin/contacts/mailto.html"

    def _get_liste_destinataires(self, ct):
        """
        Retourne la liste des adresses électroniques des objects sélectionnés
        ainsi que de leurs contacts associés (sauf s'ils ne souhaitent pas
        recevoir de courriel automatique) lus en session dans le « panier » de
        sélection, selon le content type donné en paramètre.
        Deux listes sont retournées:
         - la liste des adresses d'expéditeur
         - la liste des contacts sélectionnés mais ne disposant pas d'adresse.
        """
        adresses = []
        adresses_nospam = []
        sans_adresse = []
        try:
            selection = self.request.session[ct.model_class().basket_name]
        except KeyError:
            selection = []
        object_list = []
        for ct_id, oid in selection:
            contact = ContentType.objects.get(pk=ct_id)\
                            .get_object_for_this_type(pk=oid)
            liste_adr = contact.get_courriels()
            if liste_adr:
                # on collecte l'adresse si elle n'est pas déjà sélectionnée
                adrs = {}
                for nom, email in liste_adr:
                    if not adrs.has_key(email):
                        adrs[email] = True
                        ##
                        ext_mail = re.sub('  *',' ',
                                          email.replace(',',' ').replace(';',' ').strip()
                                          ).split(' ')
                        for sub_mail in ext_mail:
                            if nom: dest = u'"%s" <%s>' % (nom.strip(), sub_mail)
                            else:   dest = email
                            adresses.append(dest)
                        ##
##                        if nom: dest = u'"%s" <%s>' % (nom, email.strip())
##                        else:   dest = email
##                        adresses.append(dest)
            else:
                sans_adresse.append((unicode(contact),
                                     contact.get_absolute_url()))
        return (adresses, sans_adresse)

    def get_context_data(self, **kwargs):
        """
        Peuple le contexte du template.
        """
        context = super(GenMailto, self).get_context_data(**kwargs)
        ct = get_object_or_404(ContentType, pk=self.kwargs['ct_id'])
        destinataires, sans_adresse = self._get_liste_destinataires(ct)
        mc = ct.model_class()
        url_changelist = reverse("admin:%s_%s_changelist" \
                                 % (ct.app_label, ct.model.lower()))
        context.update({
            'destinataires': destinataires,
            'sans_adresse': sans_adresse,
            'nom_objet': { # utilisé uniquement pour le lien de retour
                'singulier': mc._meta.verbose_name,
                'pluriel': mc._meta.verbose_name_plural,
                },
            'css_app': settings.STATIC_URL+"css/contacts.css",
            'url_retour': url_changelist,
            })
        return context



### ---------------------------------------------------------------------------
# Vue abstraite de base pour la génération de CSV
#

class GenCSV(generic_views.base.View):
    """
    Vue générant un fichier CSV exportant des attributs des contacts
    sélectionnés à l'aide d'une classe désignée par classe_export.
    """
    # La classe à utiliser pour générer le CSV
    # À préciser obligatoirement dans les classes filles.
    classe_export = None
    # Le nom du fichier généré proposé à l'utilisateur
    nom_fichier = "export.csv"
    # Message affiché lorsqu'aucun objet correspondant n'est trouvé
    msg_aucun_objet = u"Aucun objet correspondant n'a été trouvé"\
                        u" dans le panier"
    # Inclure les personnes associee au structure (cf: courrier_structure)
    inclure_personne_associee = False


    def get(self, request, *args, **kwargs):
        """
        """
        # Lecture du queryset
        ct = get_object_or_404(ContentType, pk=int(self.kwargs['ct_id']))
        model_class = ct.model_class()
        try:
            selection = request.session[model_class.basket_name]
        except KeyError:
            selection = []
        object_list = []
        for ct_id, oid in selection:
            obj = ContentType.objects.get(pk=ct_id)\
                               .get_object_for_this_type(pk=oid)
            if self._objet_est_selectionne(obj):
                object_list.append(obj)

                # si inclure_personne_associee = True, on ajoute les personnes
                # associee aux structure dont la case "courrier_structure" est cochee
                # cf code GenEtiquettes.get() (ticket #2041)
                if self.inclure_personne_associee:
                    for pers_associees in obj.liste_contacts_associes():
                        if pers_associees.courrier_structure:
                            object_list.append(pers_associees)


        if len(object_list) == 0:
            messages.error(request, self.msg_aucun_objet)
            return http.HttpResponseRedirect(reverse('admin:affiche_panier',
                                       args=[ int(self.kwargs['ct_id']) ]))
        # Génération et renvoi du document
        buffer = StringIO()
        csvdoc = self.classe_export(buffer)
        csvdoc.generer(object_list)
        response = http.HttpResponse(mimetype='text/csv')
        response['Content-Disposition'] = \
                        'attachment; filename='+self.nom_fichier
        response.write(buffer.getvalue())
        buffer.close()
        return response


    def _objet_est_selectionne(self, obj):
        """
        Retourne vrai si l'objet donné doit être exporté, faux sinon
        (permet de ne sélectionner que les objets d'un certain type
        dans les classes filles).
        Par défaut, aucun filtrage n'est effectué.
        """
        return True


### ---------------------------------------------------------------------------
# Vue de génération de CSV complet
#

class GenCSVComplet(GenCSV):
    """
    Vue générant un fichier CSV exportant les attributs des contacts
    sélectionnés via la classe contacts_utils.ExportCSVComplet (qui exporte les
    champs décrits dans la liste « attrs_export » du modèle correspondant).
    """
    # La classe à utiliser pour générer le CSV
    classe_export = contacts_utils.ExportCSVComplet
    # Le nom du fichier généré proposé à l'utilisateur
    nom_fichier = "export_donnees.csv"



class GenCSVCompletStructures(GenCSVComplet):
    """
    Vue n'exportant que les objets héritant de Structure.
    """
    msg_aucun_objet = u"Aucune structure ou agriculteur n'a été trouvé"\
                        u" dans le panier"
    
    def _objet_est_selectionne(self, obj):
        """
        Retourne vrai si l'objet donné doit être exporté, faux sinon.
        Ici, on ne sélectionne que les objets héritant de Structure.
        """
        return isinstance(obj, contacts_models.Structure)


class GenCSVCompletPersonnes(GenCSVComplet):
    """
    Vue n'exportant que les objets héritant de Personnes.
    """
    msg_aucun_objet = u"Aucune personne n'a été trouvée dans le panier"
    
    def _objet_est_selectionne(self, obj):
        """
        Retourne vrai si l'objet donné doit être exporté, faux sinon.
        Ici, on ne sélectionne que les objets du modèle Personne.
        """
        return isinstance(obj, contacts_models.Personne)




### ---------------------------------------------------------------------------
# Vue de génération de CSV pour publipostage
#

class GenCSVPublipostage(GenCSV):
    """
    Vue générant un fichier CSV exportant les attributs nécessaires à un
    publipostage via la classe contacts_utils.ExportCSVPublipostage (qui
    exporte les champs désignés par l'attribut « attrs_publipostage » du
    modèle correspondant).
    """
    inclure_personne_associee = True
    # La classe à utiliser pour générer le CSV
    classe_export = contacts_utils.ExportCSVPublipostage
    # Le nom du fichier généré proposé à l'utilisateur
    nom_fichier = "publipostage_contacts.csv"



### ---------------------------------------------------------------------------
# Vue créant un nouvel objet DonneesBio pour un agriculteur bio et redirigant
# vers le formulaire de modification de cet objet.
# (évite l'utilisation de la longue liste des agriculteurs bio).
#

class CreeDonneesBio(generic_views.base.RedirectView):
    """
    """

    def get(self, request, *args, **kwargs):
        """
        """
        agri_bio = contacts_models.AgriculteurBio.objects.get(
                                pk=self.kwargs['agriculteurbio_id'])
        self.donneesbio = contacts_models.DonneesBio.objects\
                                .create(agriculteur_bio=agri_bio)
        return generic_views.base.RedirectView.get(self, request,
                                                   *args, **kwargs)

    def get_redirect_url(self, **kwargs):
        """
        Retourne l'URL de la vue de modification de l'objet DonneesBio
        récemment créé.
        """
        return self.donneesbio.get_absolute_url()



### ---------------------------------------------------------------------------
# Vue transformant un agriculteur en agriculteur Bio et redirigeant vers sa
# fiche d'agriculteur Bio.
#

class TransformeAgriculteur(generic_views.base.RedirectView):
    """
    """
    def get_redirect_url(self, **kwargs):
        """
        Transforme l'agriculteur donné en agriculteurbio et retourne l'URL de
        la vue de modification de l'objet AgriculteurBio créé.
        """
        agri = contacts_models.Agriculteur.objects.get(
                                pk=self.kwargs['agriculteur_id'])
        contacts_utils.transforme_agriculteur(agri)
        return reverse('admin:%s_agriculteurbio_change' % agri._meta.app_label,
                       args=[ agri.id ])


#
# Fin
### ---------------------------------------------------------------------------
