# -*- coding: utf-8 -*-

import datetime
from decimal import Decimal

from django.conf import settings
from django.contrib import admin

import django.forms as forms
import django.db.models as models
import django.contrib.auth.models as auth_models

import basecontacts_gabnor.references.models as ref

from basecontacts_gabnor import logger


### --------------------------------------------------------------------------
# Données.
# Ces données sont utilisées surtout dans models.py, mais aussi dans forms.py.
# On ne peut pas importer models.py depuis forms.py à cause d'une référence
# circulaire, on met donc ces données ici, même si ce n'est pas le plus
# logique.
#


CHOIX_TYPES_RDV = (
        (u'premier_diagnostic','Premier diagnostic'),
        (u'etude', 'Étude'),
        (u'suivi_conversion', 'Suivi de conversion'),
        (u'visite_annuelle', 'Visite annuelle'),
        )

CHOIX_MODES_COMMERCIALISATION = (
        (u'biocabas', 'Biocabas'),
        (u'vente_direct', 'Vente directe'),
        (u'AMAP', 'AMAP'),
        )

CHOIX_ABONNEMENT_LETTRE_ADHERENTS = (
        (u'non', 'Non abonné'),
        (u'courriel', 'Par courrier électronique'),
        (u'courier', 'Par courrier papier'),
        )

# Choix de l'année de premier contact
CHOIX_ANNEE = [ (i, i) for i in range(2008, 2019) ]


# Choix de mois (de premier contact)
CHOIX_MOIS = (
        (1, 'Janvier'),
        (2, 'Février'),
        (3, 'Mars'),
        (4, 'Avril'),
        (5, 'Mai'),
        (6, 'Juin'),
        (7, 'Juillet'),
        (8, 'Août'),
        (9, 'Septembre'),
        (10, 'Octobre'),
        (11, 'Novembre'),
        (12, 'Décembre'),
        )


### ---------------------------------------------------------------------------
# Données et fonctions utilitaires
#


CHOIX_NULLBOOLEAN = (('ind','Indifférent'), ('oui','Oui'), ('non', 'Non'))

def coerce_radionullbool(val, inverse=False):
    """
    Fonction accessoire utilisée par les TypedChoiceField affichant un boolean
    pouvant être nul, sous forme de boutons radio (plus ergonomique que le
    selection utilisé par défaut par NullBooleanField).
    Si inverse=True, on retourne True pour oui, pour pouvoir utiliser le
    booleen avec une requête SQL qui utilise __isnull (qui doit être True
    lorsqu'il n'y a *pas* d'objet en relation).
    """
    if val == 'oui':
        return not inverse
    if val == 'non':
        return inverse
    return None

def coerce_radionullbool_inverse(val):
    """
    Similaire à coerce_radionullbool, mais renvoit False pour oui et True pour
    non pour être utilisé directement dans une requête __isnull qui doit être
    True quand il n'y a pas d'objet en relation (ex: pas d'adhésion gabnor pour
    un agriculteur bio).
    """
    return coerce_radionullbool(val, inverse=True)



### ---------------------------------------------------------------------------
# Form de filtres spéciaux des listes
#


class FormFiltre(forms.Form):
    """
    Form de base pour filtrer une liste d'objets dans l'admin dango selon des
    filtres avancés (contrairement aux filtres fournis par django, on peut
    sélectionner plusieurs valeurs, un 'OU' étant appliquées entre ces
    valeurs).
    """
    class Media:
        css = {
            # forms.css nécessaire pour le style des filtres spéciaux
            "all": (settings.ADMIN_MEDIA_PREFIX+"css/forms.css",
                    settings.STATIC_URL+"css/contacts_forms.css"),
        }
        js = ( settings.ADMIN_MEDIA_PREFIX+"js/collapse.min.js", )

    # correspondance entre champs et argument de filtre du queryset
    filtres_qs = {}

    def construit_filtres_qs(self, cleaned_data):
        """
        Retourne une liste d'objets Q pour filtrer un queryset à partir des
        données du formulaire (qui sont conservées en session, c'est pourquoi
        on les passe en argument).
        (cette méthode pourrait aussi bien figurer dans la vue -- dans
        contacts_admin.MixinAdminFiltrable --, mais elle dépend aussi fortement
        des données du formulaire filtres_qs, on peut donc la considérer comme
        liée au formulaire.)
        """
        q_list = []
        for f in self.filtres_qs.keys():
            f_val = cleaned_data.get(f, [])
            #logger.debug("  valeur %s : %s" % (f, f_val))
            if isinstance(f_val, models.query.QuerySet):
                q_val = [ o.id for o in f_val ]
            elif isinstance(f_val, list):
                if len(f_val) == 0:
                    # un QuerySet vide apparaît comme une liste
                    continue
                q_val = f_val
            elif isinstance(f_val, bool) \
                    or isinstance(f_val, Decimal) \
                    or isinstance(f_val, datetime.date):
                q_val = f_val
            else:
                # S'il s'agit d'autre chose, c'est un filtre vide
                # qu'on ignore (une date non choisie vaut None, etc.)
                continue
            logger.debug("+ filtre: %s" % { self.filtres_qs[f]: q_val })
            q_list.append(models.Q(**{ self.filtres_qs[f]: q_val }))
        return q_list



### ---------------------------------------------------------------------------
#
#


class FormFiltreStructure(FormFiltre):
    """
    Formulaire pour filtrer la liste des structures.
    """
    filtres_qs = dict(
            etiquettes="etiquettes__id__in",
            categorie="categorie__id__in",
            zone="commune__zone__id__in",
            commune_enjeu_eau="commune__enjeu_eau",
            )

    etiquettes = forms.ModelMultipleChoiceField(
            required=False, queryset=ref.Etiquette.objects.all(),
            widget=forms.CheckboxSelectMultiple,
            label="Étiquettes")

    categorie = forms.ModelMultipleChoiceField(
            required=False, queryset=ref.CategorieStructure.objects.all(),
            widget=forms.CheckboxSelectMultiple,
            label="Catégorie")

    zone = forms.ModelMultipleChoiceField(
            required=False, queryset=ref.Zone.objects.all(),
            widget=forms.CheckboxSelectMultiple,
            label="Zone")

    commune_enjeu_eau = forms.TypedChoiceField(
            required=False, choices=CHOIX_NULLBOOLEAN,
            initial='ind',
            coerce=coerce_radionullbool,
            widget=forms.RadioSelect,
            label="Sur commune à enjeu eau ?")



class FormFiltreAgriculteur(FormFiltreStructure):
    """
    Formulaire pour filtrer la listes des agriculteurs.
    """
    filtres_qs = dict(
            FormFiltreStructure.filtres_qs,
            type_activite="type_activite__id__in",
            qualificatifs_complementaires="qualificatifs_complementaires__id__in",
            type_rendez_vous="projetbio__rendezvousprojetbio__type__in",
            technicien_suivi="projetbio__technicien__id__in",
            technicien_rdv="projetbio__rendezvousprojetbio__technicien__id__in",
            date_rdv_anterieure="projetbio__rendezvousprojetbio__date__gte",
            date_rdv_posterieure="projetbio__rendezvousprojetbio__date__lte",
            )

    type_activite = forms.ModelMultipleChoiceField(
            required=False, queryset=ref.TypeActivite.objects.all(),
            widget=forms.CheckboxSelectMultiple,
            label="Type d'activité")

    qualificatifs_complementaires = forms.ModelMultipleChoiceField(
            required=False,
            queryset=ref.AutreQualificatifExploitation.objects.all(),
            widget=forms.CheckboxSelectMultiple,
            label="Autres qualificatifs de l'exploitation")

    technicien_suivi = forms.ModelMultipleChoiceField(
            required=False,
            queryset=auth_models.User.objects.exclude(username='admin'),
            widget=forms.CheckboxSelectMultiple,
            label="Ayant au moins un projet bio suivi par")

    type_rendez_vous = forms.MultipleChoiceField(
            required=False, choices=CHOIX_TYPES_RDV,
            widget=forms.CheckboxSelectMultiple,
            label="Ayant au moins un rendez-vous pour projet bio de type")

    technicien_rdv = forms.ModelMultipleChoiceField(
            required=False,
            queryset=auth_models.User.objects.exclude(username='admin'),
            widget=forms.CheckboxSelectMultiple,
            label="Ayant au moins un rendez-vous pour projet bio effectué par")

    date_rdv_anterieure = forms.DateField(
            required=False,
            widget=admin.widgets.AdminDateWidget,
            label="Ayant au moins un rendez-vous pour projet bio après le")

    date_rdv_posterieure = forms.DateField(
            required=False, widget=admin.widgets.AdminDateWidget,
            label="Ayant au moins un rendez-vous pour projet bio avant le")

    def clean(self):
        """
        Vérification des contraintes sur les valeurs de filtrages.
        """
        d = super(FormFiltreAgriculteur, self).clean()
        rdv_min = d['date_rdv_anterieure']
        rdv_max = d['date_rdv_posterieure']
        if rdv_min is not None \
                and rdv_max is not None \
                and rdv_min > rdv_max:
            raise forms.ValidationError("la date de rendez-vous minimale"
                    " est postérieure à la date maximale")
        return d



class FormFiltreAgriculteurBio(FormFiltreAgriculteur):
    """
    Formulaire pour filtrer la liste des agriculteurs bio.
    """
    filtres_qs = dict(
        FormFiltreAgriculteur.filtres_qs,
        type_activite="type_activite__id__in",
        adherent_gabnor="adhesiongabnor__isnull",
        montant_adhesion_gabnor_min="adhesiongabnor__montant__gte",
        montant_adhesion_gabnor_max="adhesiongabnor__montant__lte",
        date_adhesion_anterieure="premiere_adhesion__date__gte",
        date_adhesion_posterieure="premiere_adhesion__date__lte",
        date_renouvellement_anterieure="renouvellement_adhesion__date__gte",
        date_renouvellement_posterieure="renouvellement_adhesion__date__lte",
        )
    adherent_gabnor = forms.TypedChoiceField(
            required=False, choices=CHOIX_NULLBOOLEAN, initial='ind',
            coerce=coerce_radionullbool_inverse,
            widget=forms.RadioSelect,
            label="Adhérent GABnor ?")

    montant_adhesion_gabnor_min = forms.DecimalField(
            required=False,
            label="Montant minimum de l'adhésion GABnor (en €)")

    montant_adhesion_gabnor_max = forms.DecimalField(
            required=False,
            label="Montant maximum de l'adhésion GABnor (en €)")

    date_adhesion_anterieure = forms.DateField(
            required=False, widget=admin.widgets.AdminDateWidget,
            label="Ayant adhéré après le")

    date_adhesion_posterieure = forms.DateField(
            required=False, widget=admin.widgets.AdminDateWidget,
            label="Ayant adhéré avant le")

    date_renouvellement_anterieure = forms.DateField(
            required=False, widget=admin.widgets.AdminDateWidget,
            label="Ayant renouvelé son adhésion après le")

    date_renouvellement_posterieure = forms.DateField(
            required=False, widget=admin.widgets.AdminDateWidget,
            label="Ayant renouvelé son adhésion avant le")

    def clean(self):
        """
        Vérification des contraintes sur les valeurs de filtrages.
        """
        d = super(FormFiltreAgriculteurBio, self).clean()
        montant_min = d['montant_adhesion_gabnor_min']
        montant_max = d['montant_adhesion_gabnor_max']
        if montant_min is not None \
                and montant_max is not None \
                and montant_min > montant_max:
            raise forms.ValidationError("le montant minimal d'adhésion"
                    " est postérieure au montant maximal")
        adhesion_min = d['date_adhesion_anterieure']
        adhesion_max = d['date_adhesion_posterieure']
        if adhesion_min is not None \
                and adhesion_max is not None \
                and adhesion_min > adhesion_max:
            raise forms.ValidationError("la date d'adhésion minimale"
                    " est postérieure à la date maximale")
        renouvellement_min = d['date_renouvellement_anterieure']
        renouvellement_max = d['date_renouvellement_posterieure']
        if renouvellement_min is not None \
                and renouvellement_max is not None \
                and renouvellement_min > renouvellement_max:
            raise forms.ValidationError("la date de renouvellement minimale"
                    " est postérieure à la date maximale")
        return d



### ---------------------------------------------------------------------------
# Form pour les actions sur le panier
#


class BaseFormActionPanier(forms.Form):
    """
    Formulaire pour le choix des actions à effectuer sur la sélection.
    Il s'agit d'une classe abstraite, dont les filles doivent définir
    CHOIX_ACTION.
    """
    class Media:
        css = {
            "all": (settings.ADMIN_MEDIA_PREFIX+"css/changelists.css",
                    settings.STATIC_URL+"css/contacts.css",),
        }
    
    CHOIX_ACTION = ()
    action = forms.ChoiceField(choices=CHOIX_ACTION) 

    def __init__(self, *args, **kwargs):
        """
        Modifie action.choices en utilisant le CHOIX_ACTION de l'objet
        (donc celui de la classe fille).
        """
        self.base_fields['action'].choices = self.CHOIX_ACTION
        super(BaseFormActionPanier, self).__init__(*args, **kwargs)


class FormActionPanierContact(BaseFormActionPanier):
    """
    Formulaire pour choix d'actions pour les modèles représentant un contact
    (avec les actions associées à ce type de données).
    """
    CHOIX_ACTION = (
            ('', '---'),
            ('gen_etiquettes', u'Générer des étiquettes'),
            ('gen_mailto', u'Envoyer un courriel'),
            ('export_csv_publipostage', u'Exporter pour publipostage'),
            ('export_csv_complet_structures',
                u'Exporter les structures de la sélection en CSV'),
            ('export_csv_complet_personnes',
                u'Exporter les personnes de la sélection en CSV'),
            )


class FormActionPanierDonnee(BaseFormActionPanier):
    """
    Formulaire pour choix d'actions pour les modèles représentant des données
    (avec les actions associées à ce type de données).
    """
    CHOIX_ACTION = (
            ('', '---'),
            (u'export_csv_complet', u'Exporter les données en CSV'),
            )



#
### ---------------------------------------------------------------------------
