# -*- coding: utf-8 -*-
# contacts.models

import datetime
from django.db import models
from django.core import exceptions
from django.core.urlresolvers import reverse
import django.contrib.auth.models as auth_models

import basecontacts_gabnor.references.models as ref
import basecontacts_gabnor.contacts.forms as contacts_forms

from basecontacts_gabnor import logger

LOCAL_DATE_FORMAT = '%d/%m/%Y'


### Données ------------------------------------------------------------------
#

CHOIX_CIVILITE = (
        (u'M', 'Monsieur'),
        (u'Mme', 'Madame'),
        (u'Melle', 'Mademoiselle'),
        )

CHOIX_PARCS_NATURELS_REGIONAUX = (
        (u"", "Aucun"),
        (u"Parc naturel régional des Caps et Marais d'Opale",
         "Parc naturel régional des Caps et Marais d'Opale"),
        (u"Parc naturel régional Scarpe-Escaut",
         "Parc naturel régional Scarpe-Escaut"),
        (u"Parc naturel régional de l'Avesnois",
         "Parc naturel régional de l'Avesnois"),
        )

CHOIX_VISIBILITE = (
        (u"personne",    u"Dépublié"),
        (u"tous",        u"Visible par tous"),
        (u"adherents",   u"Visible par les adhérents seulement"),
        )

# Type de vente (utilisé dans Assolement)
CHOIX_TYPE_CIRCUIT = (
        (u"VD", "Vente Directe"),
        (u"CC", "Circuit court"),
        (u"CL", "Circuit long"),
        )

CHOIX_COMMERCIALISATION_VENTE_DIRECTE = (
        (u"paniers", "Paniers"),
        (u"vente à la ferme", "Vente à la ferme"),
        (u"marchés", "Marchés"),
        (u"autres", "Autres"),
        )

CHOIX_COMMERCIALISATION_CIRCUIT_COURT = (
        (u"biocabas", "Biocabas"),
        (u"distributeurs", "Distributeurs"),
        (u"autres", "Autres"),
        )

CHOIX_COMMERCIALISATION_CIRCUIT_LONG = (
        (u"transformateurs", u"Transformateurs"),
        (u"marchés de gros", u"Marché de gros"),
        (u"autres", u"Autres"),
        )

CHOIX_SYSTEME_FORCAGE_ENDIVES = (
        (u"aucun", "Aucun"),
        (u"en salle", "En salle"),
        (u"en pleine terre", "En pleine terre"),
        )

CHOIX_TYPE_ATELIER = (
        (u"dispense d'agrément", u"Dispense d'agrément"),
        (u"agrément CEE", u"Agrément CEE"),
        )

### --------------------------------------------------------------------------
# Exceptions de l'application
#

class AppImproperlyConfigured(exceptions.ImproperlyConfigured):
    """L'application est mal configurée."""



### --------------------------------------------------------------------------
# Classes principales
#


class Contact(object):
    """
    Classe d'interface définissant l'API commune entre les modèles Structure et
    Personne qui sont parfois traités de façon identique puisqu'ils partagent
    le même panier.
    """
    # Le panier où conserver les objets sélectionnés
    basket_name = u"contacts"

    # Form fournissant les actions possibles sur le panier
    classe_form_actions_panier = contacts_forms.FormActionPanierContact

    # Champs à inclure dans un export CSV pour publipostage
    attrs_publipostage = [
            'get_designation', 'adresse', 'code_postal', 'ville',
            'tel_fixe', 'tel_mobile', 'fax', 'courriel', ]

    def get_designation(self):
        """
        Retourne le nom de la structure ou de la première personne associée.
        (utilisée lors de l'exportation CSV pour publipostage)
        """
        return unicode(self)

    get_designation.verbose_name = u"Désignation ou personne"

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresses email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """

    def get_courriels(self):
        """
        Retourne le courriel de la structure et des personnes associées
        (sauf s'ils ne souhaitent pas recevoir d'email automatique)
        """

    def liste_contacts_associes(self, courrier=False):
        """
        Méthode commune aux modèles Structure et Personne pour un traitement
        indifférencié de ces objets. Retourne la liste des contacts associés à
        cet objet. Pour un objet Personne, il n'y en a pas.
        Si courrier vaut True, retourne uniquement ceux devant recevoir un
        courrier destiné à la structure (utilisé uniquement dans Structure).
        """
        return []
    



class Structure(Contact, models.Model):
    u"""
    Structure générique: utilisé pour les structures « génériques », les
    agriculteurs, et agriculteurs bio.
    .
    Attributs et méthodes hérités de la classe abstraite Contact (hérités aussi
    par Personne) pour permettre le partage du même panier avec le modèle
    Personne.
     - basket_name
     - classe_form_actions_panier
     - attrs_publipostage
     - adresse_electronique()
     - get_courriels()
    """
    class Meta:
        ordering = ['designation',]
        verbose_name = "structure"
        verbose_name_plural = "structures"

    # Champs à inclure dans un export CSV complet
    attrs_export = [
            'id', 'get_designation', 'siret', 'statut', 'categorie', 'adresse', 'ville',
            'code_postal', 'commune', 'tel_fixe', 'tel_mobile', 'fax',
            'courriel', 'pas_de_courriel',
            'site_web', 'date_creation', 'date_derniere_modification',
            'utilisateur_derniere_modification', 'origine_contact',
            'annee_contact', 'mois_contact', 'sage', 'commentaire',
            'etiquettes',
            # Attributs propres au modèle Agriculteur
            'visibilite', 'type_activite', 'qualificatifs_complementaires',
            'quota_laitier', 'laiterie', 'nb_vaches_allaitantes',
            'sau', 'sau_bio', 'sau_eau', 'sau_eau_bio',
            'parc_naturel_regional',
            # Attributs propres au modèle AgriculteurBio
            'date_envoi_courrier', 'premiere_adhesion',
            'renouvellement_adhesion', 'get_date_derniere_adhesion',
            'get_date_premiere_conversion',
            'type'
            ]
    # Modèle d'où tirer les verbose_name des champs exportés
    # (si l'attribut n'est pas défini, c'est le modèle lui-même)
    attrs_verbose_from = "AgriculteurBio"

    designation = models.CharField(max_length=250, verbose_name=u"désignation")
    statut = models.ForeignKey(ref.StatutStructure, null=True, blank=True)
    categorie = models.ForeignKey(ref.CategorieStructure, null=True, blank=True,
                                  verbose_name=u"catégorie")
    adresse = models.TextField(null=True, blank=True)
    commune = models.ForeignKey(ref.Commune, null=True, blank=True)
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(max_length=10, null=True, blank=True,
           help_text=u"Laisser vide pour utiliser celui associé à la commune")
    # la ville contient également le code cedex (à utilser pour l'adresse)
    ville = models.CharField(max_length="255", null=True, blank=True,
           help_text=u"Nom de ville avec code cédex si nécessaire, "\
                     u" si différent du nom de la commune.")
    tel_fixe = models.CharField(max_length=100, null=True, blank=True,
                                verbose_name=u"téléphone fixe")
    tel_mobile = models.CharField(max_length=100, null=True, blank=True,
                                  verbose_name=u"téléphone mobile")
    fax = models.CharField(max_length=100, null=True, blank=True)
    courriel = models.CharField(max_length=250, null=True, blank=True)
    siret = models.CharField(max_length=14, null=True, blank=True,
                             verbose_name=u"N. Siret", help_text="14 chiffres sans espaces")
    pas_de_courriel = models.BooleanField(default=False,
            verbose_name=u"pas d'envoi automatique de courriel",
            help_text=u"Cochez cette case si cette structure ne souhaite"
                      u" pas recevoir de courriel envoyé automatiquement.")
    site_web = models.URLField(max_length=250, null=True, blank=True,
                               verbose_name=u"URL du site web")
    date_creation = models.DateField(editable=False,
            verbose_name=u"date de saisie de l'enregistrement")
    date_derniere_modification = models.DateField(editable=False,
            verbose_name=u"date de dernière modification des informations")
    utilisateur_derniere_modification = models.ForeignKey(auth_models.User,
            editable=False, verbose_name=u"dernière modification par")
    origine_contact = models.CharField(max_length=255, null=True, blank=True,
                                       verbose_name=u"origine du contact")
    annee_contact = models.IntegerField(null=True, blank=True,
                        choices=contacts_forms.CHOIX_ANNEE,
                        verbose_name=u"Année du premier contact")
    mois_contact = models.IntegerField(null=True, blank=True,
                        choices=contacts_forms.CHOIX_MOIS,
                        verbose_name=u"Mois du premier contact")
    sage = models.CharField(max_length=255, null=True, blank=True,
                            verbose_name=u"SAGE")
    commentaire = models.TextField(null=True, blank=True)
    etiquettes = models.ManyToManyField(ref.Etiquette, null=True, blank=True,
                                        verbose_name=u'étiquettes')
    
    def __unicode__(self):
        """
        Retourne une représentation unicode de la structure.
        """
        if self.designation:
            return self.designation
        if self.personne_set.count() > 0:
            return unicode(self.personne_set.all()[0])
        return u"Structure (#%d)" % self.id

    def get_absolute_url(self):
        """
        Retourne l'URL associée à l'objet.
        """
        return reverse(
                'admin:%s_%s_change' % (self._meta.app_label,
                                        self.__class__.__name__.lower()),
                args=[self.pk])

    def contenu_etiquette(self):
        """
        Retourne le contenu à afficher dans l'étiquette sous forme de liste
        dont chaque élément représente une ligne.
        """
        contenu = []
        # On doit remplacer & par &amp; à cause d'un problème dans reportlab
        # qui génère une erreur XML lorsque le nom contient un "&"
        contenu.append(unicode(self).replace("&", "&amp;"))
        contenu.extend(self.adresse.split('\n'))
        contenu.append(u"%s %s" % (self.code_postal or '', self.ville or ''))
        return contenu

    def liste_contacts_associes(self):
        """
        Surcharge de la méthode de Contact.
        Retourne la liste des contacts associés à cet objet. Il
        s'agit ici des personnes associées à la structure.
        """
        return self.personne_set.all()

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresse email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        return (unicode(self), self.courriel)

    def get_courriels(self):
        """
        Retourne le courriel de la structure et des personnes associées
        (sauf s'ils ne souhaitent pas recevoir d'email automatique)
        """
        courriels = []
        if self.courriel:
            courriels.append(self.adresse_electronique())
        for p in self.liste_contacts_associes():
            courriels.extend(p.get_courriels())
        return courriels

    def type(self):
        """
        retourne le type parmi structure, agri, agribio
        """
#        return "structure"
#        return self.__class__.__name__
        try:
            self.agriculteur
        except Agriculteur.DoesNotExist:
            return "structure"
        try:
            self.agriculteur.agriculteurbio
        except AgriculteurBio.DoesNotExist:
            return "agri"
        return "bio"
    

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse, mieux
        vaut le gérer manuellement).
        """
        if self.id is None and self.date_creation is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        models.Model.save(self, *args, **kwargs)



#class StructureGenerique(Structure):
#    """
#    Contact représentant une structure quelconque qui n'est pas Agriculteur ou
#    AgriculteurBio.
#    Ce modèle permet d'isoler les structures qui ne sont pas Agriculteur ou
#    AgriculteurBio pour ne pas les mélanger dans l'admin.
#    # MODIFIÉ: le Gabnor souhaite le comportement inverse: avoir voir toutes
#    # les structures sous « Structures », même les agriculteurs.
#    # Ce modèle est conservé pour documenter l'astuce, en cas de besoin :)
#    """
#    class Meta:
#        verbose_name = "structure générique"
#        verbose_name_plural = "structures génériques"



class Agriculteur(Structure):
    """
    Contact représentant un agriculteur.
    """
    class Meta:
        # ordering est hérité de Structure
        verbose_name = "agriculteur"
        verbose_name_plural = "agriculteurs"

#    # Champs à inclure dans un export CSV complet
    attrs_export = Structure.attrs_export \
#            + [ 'visibilite', 'type_activite', 'qualificatifs_complementaires',
#                'quota_laitier', 'laiterie', 'nb_vaches_allaitantes',
#                'sau', 'sau_bio', 'sau_eau', 'sau_eau_bio', 

    visibilite = models.CharField(max_length=255,
                    choices=CHOIX_VISIBILITE, default="personne",
                    verbose_name=u"publication sur le site gabnor.org")
    type_activite = models.ForeignKey(ref.TypeActivite, null=True, blank=True)
    qualificatifs_complementaires = \
                models.ManyToManyField(ref.AutreQualificatifExploitation,
                    null=True, blank=True,
                    verbose_name=u"autres qualificatifs de l'exploitation")
    # si qualificatifs_complementaires inclut Bovins/Lait:
    quota_laitier = models.IntegerField(null=True, blank=True,
                                verbose_name=u"quota laitier (en litres)")
    # si qualificatifs_complementaires inclut Bovins/Lait:
    laiterie = models.ForeignKey(ref.Laiterie, null=True, blank=True)
    # si qualificatifs_complementaires inclut Bovins/Viande:
    nb_vaches_allaitantes = models.CharField(null=True, blank=True,
                                max_length=100,
                                verbose_name=u"nombre de vaches allaitantes",
                                help_text='Indiquez ici un nombre (« 15 »)'
                                          ' ou une fourchette (« 10-20 »)')
    sau = models.DecimalField(max_digits=10, decimal_places=2,
                              null=True, blank=True,
                              verbose_name=u"Surface Agricole Utile",
                              help_text='Surface Agricole Utile (en ha)')
    sau_bio = models.DecimalField(max_digits=10, decimal_places=2,
                        null=True, blank=True,
                        verbose_name=u"SAU bio (en ha)",
                        help_text='Surface agricole utile biologique (en ha)')
    # si commune.enjeu_eau
    sau_eau = models.DecimalField(max_digits=10, decimal_places=2,
                          null=True, blank=True,
                          verbose_name=u"SAU à enjeu eau",
                          help_text='Surface agricole utile'
                                    ' sur commune à enjeu eau (en ha)')
    # si commune.enjeu_eau
    sau_eau_bio = models.DecimalField(max_digits=10, decimal_places=2,
                          null=True, blank=True,
                          verbose_name=u"SAU en biologique à enjeu eau",
                          help_text='Surface agricole utile en biologique'
                                    ' sur commune à enjeu eau (en ha)')
    parc_naturel_regional = models.CharField(max_length=200,
                                    null=True, blank=True,
                                    choices=CHOIX_PARCS_NATURELS_REGIONAUX,
                                    verbose_name=u"parc naturel régional")

    def est_bio(self):
        """
        Return True si l'agriculteur est aussi bio, faux sinon.
        """
        try:
            self.agriculteurbio
        except AgriculteurBio.DoesNotExist:
            return False
        return True

class ProjetBio(models.Model):
    """
    Projet de passage en bio d'un agriculteur.
    Chaque agriculteur peut en avoir plusieurs, à conserver même après le
    passage en Bio.
    """
    class Meta:
        verbose_name = "projet de passage en bio"
        verbose_name_plural = "projets de passage en bio"

    # Champs à inclure dans un export CSV complet
    attrs_export = [
            'id', 'agriculteur', 'agri_id', 'technicien', 'description', 'freins_identifies',
            'suites_a_donner', 'remarques', 'etude_demandee', 'date_etude',
            'date_probable_certification', 'date_certification',
            'engagement_oc', 'engagement_oc_enjeu_eau',  'installation_bio', 'date_engagement_oc',
            'commentaire', ]

    # Le panier où conserver les objets sélectionnés
    basket_name = u"projets bio"

    # Form fournissant les actions possibles sur le panier
    classe_form_actions_panier = contacts_forms.FormActionPanierDonnee
    
    agriculteur = models.ForeignKey('Agriculteur')
    technicien = models.ForeignKey(auth_models.User, null=True, blank=True,
                                   verbose_name=u"suivi par",
                                   on_delete=models.SET_NULL)
    description = models.TextField(null=True, blank=True,
                                   verbose_name=u"description")
    freins_identifies = models.TextField(null=True, blank=True,
                                         verbose_name=u"freins identifiés")
    suites_a_donner = models.TextField(null=True, blank=True,
                                       verbose_name=u"suites à donner")
    remarques = models.TextField(null=True, blank=True,
                                 verbose_name=u"remarques")
    etude_demandee = models.NullBooleanField(u"Étude demandée ?")
    date_etude = models.DateField(u"Date de l'étude", null=True, blank=True)
    date_probable_certification = models.CharField(null=True, blank=True,
        choices=(('long terme', 'Long terme'),
                 ('moyen terme', 'Moyen terme'),
                 ('date fournie', 'Date précisée ci-dessous')),
        max_length=20, verbose_name=u"date probable de certification",
        help_text="Si vous choisissez l'option « Date précisée ci-dessous »,"
                  " merci de préciser la date dans le champ ci-dessous.")
    date_certification = models.DateField(null=True, blank=True,
            verbose_name=u"Date de certification prévue")
    engagement_oc = models.IntegerField(null=True, blank=True,
                                        verbose_name=u"Engagement OC (en ha)")
    engagement_oc_enjeu_eau = models.DecimalField(null=True, blank=True,
                                                  verbose_name=u"Engagement OC (en ha) dont enjeu eau",
                                                  help_text=u"utilisez . comme séparateur décimal",
                                                  decimal_places=4, max_digits=10)
    installation_bio = models.NullBooleanField(
                                    verbose_name=u"installation bio")
    date_engagement_oc = models.DateField(null=True, blank=True, db_index=True,
                                    verbose_name=u"date d'engagement de l'OC")
    commentaire = models.TextField(null=True, blank=True,
                                   verbose_name=u"commentaire")

    def __unicode__(self):
        """
        Retourne une représentation unicode de l'objet.
        Attention, problème vicieux: l'admin django appelle cette méthode juste
        après la suppression de l'objet dans la base: il ne faut donc pas
        utiliser %d pour self.id car dans ce cas, self.id vaut None et provoque
        une erreur.
        """
        return u"projet bio de %s (id #%s)" % (self.agriculteur, self.id)

    def agri_id(self):
        if self.agriculteur:
            return self.agriculteur.id
        return u""
    agri_id.short_description = u"ID agriculteur"
    agri_id.verbose_name = u"ID agriculteur"
    

class AgriculteurBio(Agriculteur):
    """
    Agriculteur ayant une activité biologique.
    """
    class Meta:
        verbose_name = "agriculteur biologique"
        verbose_name_plural = "agriculteurs biologiques"

#    # Champs à inclure dans un export CSV complet
#    attrs_export = Agriculteur.attrs_export \
#        + [ 'date_envoi_courrier', 'premiere_adhesion',
#            'renouvellement_adhesion', 'get_date_derniere_adhesion',
#            'get_date_premiere_conversion' ]

    date_envoi_courrier = models.DateField(null=True, blank=True,
                verbose_name=u"date d'envoi du courrier de relance adhésion")
    premiere_adhesion = models.ForeignKey(
                'AdhesionGabnor', null=True, blank=True,
                related_name="premieres_adhesions",
                verbose_name=u"Première adhésion",
                on_delete=models.SET_NULL)
    renouvellement_adhesion = models.ForeignKey(
                'AdhesionGabnor', null=True, blank=True,
                related_name="renouvellements",
                verbose_name=u"dernier renouvellement d'adhésion",
                on_delete=models.SET_NULL)
    # + champ automatique donneesbio_set

    def get_date_premiere_adhesion(self):
        """
        Retourne la date de première adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.premiere_adhesion.date
        except AdhesionGabnor.DoesNotExist:
            return None

    get_date_premiere_adhesion.verbose_name = u"date de première adhésion"

    def get_date_derniere_adhesion(self):
        """
        Retourne la date de dernière adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.adhesiongabnor_set.latest().date
        except AdhesionGabnor.DoesNotExist:
            return None

    get_date_derniere_adhesion.verbose_name = u"date de dernière adhésion"

    def get_date_premiere_conversion(self):
        """
        Retourne la date de première conversion bio, ou None s'il n'y en a pas
        (ce qui ne devrait pas être le cas pour un agriculteur bio, mais
        restons parano^Wprudent).
        """
        conv = self.projetbio_set.filter(date_engagement_oc__isnull=False)\
                                 .order_by('date_engagement_oc')
        if len(conv) == 0:
            return None
        return conv[0].date_engagement_oc

    get_date_premiere_conversion.verbose_name = u"date de première conversion"
    get_date_premiere_conversion.short_description = u"date de première conversion"

    def premiere_conversion(self):
        return self.get_date_premiere_conversion()

    def get_liste_donneesbio(self):
        """
        Retourne la listes des données de l'observatoire de l'agriculteur
        (utilisé pour afficher les liens vers ceux-ci dans la vue de
        modification de l'agriculteur).
        """
        return [
            {
                'url': dbio.get_absolute_url(),
                'annee': dbio.annee,
            }
            for dbio in self.donneesbio_set.order_by('-annee')
            ]

    def get_url_cree_donneesbio(self):
        """
        Retourne l'URL de la vue permettant de créer un nouvel objet DonneesBio
        pour cet agriculteur et d'afficher sa page de modification.
        """
        return reverse('admin:cree_donneesbio', args=[ self.id ])


class DonneesBio(models.Model):
    """
    Données de l'« observatoire » du Gabnor pour une année spécifique.
    """
    class Meta:
        verbose_name = "données annuelles de l'observatoire bio"
        verbose_name_plural = "données annuelles de l'observatoire bio"
        ordering = ['agriculteur_bio', 'annee']

    # Champs à inclure dans un export CSV complet
    attrs_export = [
        'agriculteur_bio', 'agri_id', 'annee', 'organisme_certification',
        'aide_conversion_ab', 'date_aide_conversion_ab', 'aide_maintien_ab',
        'aide_credit_impot', 'aide_certification_bio', 'aide_investissement',
        'aide_installation', 'aide_mae_t', 'aide_autres',
        'aide_accompagnement', 'aide_accompagnement_precisions',
        'mixite', 'biodynamie', 'etp_associes', 'etp_familiaux',
        'etp_salaries_permanents', 'etp_saisonniers', 'etp_encadrants',
        'personnes_insertion', 'reseaux_engagements', 'remarques',
        'vente_a_la_ferme', 'vente_ferme_detail', 'produits_dispo_marches',
        'vente_marches_detail', 'produits_dispo_biocabas', 'biocabas_detail',
        'produits_dispo_magasins', 'magasins_detail',
        'produits_dispo_restaurateurs', 'restaurateurs_detail',
        'restau_collective', 'restau_collective_detail',
        'commerc_cl', 'commerc_cl_detail',
        'statut_commercialisation', 'statut_choisi', 'types_services',
        'detail_activites', 'legumes', 'legumes_cultives_annee',
        'legumes_commercialisation_vd', 'legumes_vd_autres',
        'legumes_commercialisation_cc', 'legumes_cc_autres',
        'legumes_commercialisation_cl', 'legumes_cl_autres', 'legumes_hors_vd',
        'legumes_transformation', 'legumes_detail_transformation',
        'legumes_systeme_forcage_endives', 'culture', 'culture_comm_vd',
        'culture_comm_cc', 'culture_comm_cl', 'culture_hors_vd',
        'culture_transformation', 'culture_detail_transformation',
        'culture_stockage_ferme', 'vergers_bt', 'vergers_bt_fruits',
        'vergers_bt_comm_vd', 'vergers_bt_comm_cc', 'vergers_bt_comm_cl',
        'vergers_bt_hors_vd', 'vergers_bt_transformation',
        'vergers_bt_detail_transformation',
        'vergers_ht', 'vergers_ht_pommes', 'vergers_ht_poires',
        'vergers_ht_prunes', 'vergers_ht_comm_vd', 'vergers_ht_comm_cc',
        'vergers_ht_comm_cl', 'vergers_ht_hors_vd',
        'vergers_ht_transformation', 'vergers_ht_detail_transformation',
        'autres_petits_fruits',
        'autres_petits_fruits_detail', 'autres_plants_maraichers',
        'autres_plants_detail', 'autres_nb_plants_an', 'autres_ppam',
        'autres_ppam_detail', 'autres_prod_veg', 'autres_prod_veg_detail',
        'autres_comm_vd', 'autres_comm_cc', 'autres_comm_cl', 'autres_hors_vd',
        'autres_transformation', 'autres_detail_transformation', 'lait_vaches',
        'lait_nb_vaches', 'lait_vaches_conversion', 'lait_quota_laiterie',
        'lait_quota_vente_directe', 'lait_prod_laiterie',
        'lait_prod_vente_directe', 'lait_comm_vd', 'lait_comm_cc',
        'lait_comm_cl', 'lait_hors_vd', 'lait_laiterie',
        'lait_transformation_ferme', 'lait_transformation_presta',
        'lait_atelier_lait', 'lait_produits_transformes', 'lait_reformes_cl',
        'lait_genisses_cl', 'lait_taurillons_cl', 'lait_boeufs_cl',
        'lait_caissette', 'lait_adultes_cc', 'lait_veaux_cc',
        'lait_abattoirs_vl', 'lait_decoupe_presta', 'lait_atelier_decoupe',
        'lait_chevres', 'lait_nb_chevres', 'lait_chevres_conversion',
        'lait_chevres_prod', 'lait_chevres_transformation_ferme',
        'lait_chevres_transformation_presta', 'lait_chevres_atelier_lait',
        'lait_chevres_type_produits', 'lait_chevres_comm_vd',
        'lait_chevres_comm_cc', 'lait_chevres_comm_cl', 'lait_chevres_hors_vd',
        'bovine', 'bovine_nb_va', 'bovine_nb_conversion', 'bovine_comm_vd',
        'bovine_comm_cc', 'bovine_comm_cl', 'bovine_hors_vd',
        'bovine_reformes_cl', 'bovine_genisses_cl', 'bovine_taurillons_cl',
        'bovine_boeufs_cl', 'bovine_caissette', 'bovine_adultes_cc',
        'bovine_veaux_cc', 'bovine_abattoirs_va', 'bovine_decoupe_presta',
        'bovine_atelier_va', 'porcs', 'porcs_nb_truies', 'porcs_nb_conversion',
        'porcs_comm_vd', 'porcs_comm_cc', 'porcs_comm_cl', 'porcs_hors_vd',
        'porcs_transformation_ferme', 'porcs_produits_transformes',
        'porcs_abattoirs', 'porcs_decoupe_presta', 'porcs_atelier', 'moutons',
        'moutons_nb_brebis', 'moutons_nb_conversion', 'moutons_comm_vd',
        'moutons_comm_cc', 'moutons_comm_cl', 'moutons_hors_vd',
        'moutons_transformation_ferme', 'moutons_abattoirs',
        'moutons_decoupe_presta', 'moutons_atelier', 'animaux_lapins',
        'animaux_volailles', 'animaux_volailles_chair',
        'animaux_nb_conversion', 'animaux_nb_poulets', 'animaux_nb_lapins',
        'animaux_nb_autres_volailles', 'animaux_comm_vd', 'animaux_comm_cc',
        'animaux_comm_cl', 'animaux_hors_vd', 'animaux_atelier',
        'animaux_decoupe_presta', 'animaux_lieu_abattage',
        'animaux_autres_prod', 'animaux_details_autres_prod',
        'oeufs_nb_pondeuses', 'oeufs_nb_oeufs', 'oeufs_acheteur_cl',
        'oeufs_comm_vd', 'oeufs_comm_cc', 'oeufs_comm_cl',
        'oeufs_circuits_courts', 'miel_nb_ruches', 'conv_prod_animales',
        'conv_detail_prod_animales', 'conv_prod_vegetales',
        'conv_detail_prod_vegetales', 'conv_existe_prod_sans_valorisation',
        'conv_prod_sans_valorisation',
        'formation_annee', 'formation_annee_laquelle',
        'formation_annee_commentaire', 'formation_avenir',
        'formation_avenir_type', 'formation_avenir_details',
        # sau_agri et total_surfaces_assolement ne sont que des contrôles
        'get_assolements',
    ]

    # Le panier où conserver les objets sélectionnés
    basket_name = u"observatoires"

    # Form fournissant les actions possibles sur le panier
    classe_form_actions_panier = contacts_forms.FormActionPanierDonnee
    
    agriculteur_bio = models.ForeignKey('AgriculteurBio')
    annee = models.IntegerField(db_index=True,
                                default=datetime.date.today().year,
                                verbose_name=u"année")
    organisme_certification = models.ForeignKey(ref.OrganismeDeCertification,
                                    null=True, blank=True,
                                    on_delete=models.SET_NULL,
                                    verbose_name=u"organisme de certification")
    
    # Aides
    aide_conversion_ab = models.NullBooleanField(
            verbose_name=u"aide à la conversion à l'agriculture biologique")
    date_aide_conversion_ab = models.DateField(null=True, blank=True,
            verbose_name=u'date de cette aide',
            help_text=u"En cas d'aide à la conversion,"
                      u" indiquez la date de signature du contrat"
                      u" (ou de dépôt pour les MAE)")
    aide_maintien_ab = models.NullBooleanField(
            verbose_name=u"aide au maintien à l'agriculture biologique")
    aide_credit_impot = models.NullBooleanField(
            verbose_name=u"crédit d'impôt bio")
    aide_certification_bio = models.NullBooleanField(
            verbose_name=u"aide à la certification bio")
    aide_investissement = models.NullBooleanField(
            verbose_name=u"aide à l'investissement")
    aide_installation = models.NullBooleanField(
            verbose_name=u"aide à l'installation")
    aide_mae_t = models.NullBooleanField(
            verbose_name=u"MAE T")
    aide_autres = models.CharField(max_length=255, null=True, blank=True,
            verbose_name=u"autres mesures, précisez")
    aide_accompagnement = models.NullBooleanField(
            verbose_name=u"avez-vous besoin d'un accompagnement"
                         u" pour solliciter une aide ?")
    aide_accompagnement_precisions = models.TextField(null=True, blank=True,
            verbose_name=u"précisions")
    
    mixite = models.NullBooleanField(verbose_name=u"mixité")
    biodynamie = models.NullBooleanField()
    etp_associes = models.DecimalField(null=True, blank=True,
            max_digits=6, decimal_places=2,
            verbose_name=u"ETP d'associés")
    etp_familiaux = models.DecimalField(null=True, blank=True,
            max_digits=6, decimal_places=2,
            verbose_name=u"ETP familiaux")
    etp_salaries_permanents = models.DecimalField(null=True, blank=True,
            max_digits=6, decimal_places=2,
            verbose_name=u"ETP de salariés permanents")
    etp_saisonniers = models.DecimalField(null=True, blank=True,
            max_digits=6, decimal_places=2,
            verbose_name=u"ETP saisonniers")
    etp_encadrants = models.DecimalField(null=True, blank=True,
            max_digits=6, decimal_places=2,
            verbose_name=u"ETP encadrants")
    personnes_insertion = models.IntegerField(null=True, blank=True,
            verbose_name=u"Personnes en contrat d'insertion")
    reseaux_engagements = models.TextField(null=True, blank=True,
            verbose_name=u"Appartenance à des réseaux et engagements")
    remarques = models.TextField(null=True, blank=True,
            verbose_name=u"remarques de l'équipe")
    
    # Commercialisation
    vente_a_la_ferme = models.NullBooleanField(
                    verbose_name=u"vente à la ferme")
    vente_ferme_detail = models.TextField(null=True, blank=True,
                                verbose_name=u"détails, horaires")
    produits_dispo_marches = models.NullBooleanField(
                    verbose_name=u"marchés")
    vente_marches_detail = models.TextField(null=True, blank=True,
                    verbose_name=u"détails",
                    help_text=u"Indiquez le détail de la vente sur"
                              u" les marchés (communes, jours, horaires)")
    produits_dispo_biocabas = models.NullBooleanField(
                    verbose_name=u"biocabas")
    biocabas_detail = models.TextField(null=True, blank=True,
                verbose_name=u"détails")
    produits_dispo_magasins = models.NullBooleanField(
                    verbose_name=u"magasins")
    magasins_detail = models.TextField(null=True, blank=True,
                verbose_name=u"détails",
                help_text="Précisez la commune concernée entre parenthèses")
    produits_dispo_restaurateurs = models.NullBooleanField(
                    verbose_name=u"restaurateurs")
    restaurateurs_detail = models.TextField(null=True, blank=True,
                    verbose_name=u"détails")
    restau_collective = models.NullBooleanField(
                    verbose_name=u"restauration collective")
    restau_collective_detail = models.TextField(null=True, blank=True,
                    verbose_name=u"détails")
    commerc_cl = models.NullBooleanField(
                    verbose_name=u"circuit long (+ d'un intermédiaire)")
    commerc_cl_detail = models.TextField(null=True, blank=True,
                    verbose_name=u"détails")
    statut_commercialisation = models.NullBooleanField(
                    verbose_name=u"utilisation d'un statut spécifique"
                                 u" pour la commercialisation")
    statut_choisi = models.ForeignKey(ref.StatutCommercialisation,
                    null=True, blank=True, verbose_name=u"statut choisi",
                    help_text=u"en cas d'utilisation d'un statut spécifique"
                              u" pour la commercialisation, indiquez-le ici.")
  
    # Accueil
    types_services = models.ManyToManyField(ref.TypeService,
                        null=True, blank=True,
                        verbose_name=u"types de services")
    detail_activites = models.TextField(null=True, blank=True,
                        verbose_name=u"détail des activités")
    
    # Légumes
    legumes = models.NullBooleanField(verbose_name=u"légumes")
    legumes_cultives_annee = models.TextField(null=True, blank=True,
            verbose_name=u"légumes cultivés sur l'année")
    legumes_commercialisation_vd = models.CharField(max_length="100",
            null=True, blank=True,
            choices=CHOIX_COMMERCIALISATION_VENTE_DIRECTE,
            verbose_name=u"commercialisation en vente directe")
    legumes_vd_autres = models.CharField(
            max_length="255", null=True, blank=True,
            verbose_name=u"Autre type de vente directe")
    legumes_commercialisation_cc = models.CharField(max_length="100",
            null=True, blank=True,
            choices=CHOIX_COMMERCIALISATION_CIRCUIT_COURT,
            verbose_name=u"commercialisation en circuit court")
    legumes_cc_autres = models.CharField(
            max_length="255", null=True, blank=True,
            verbose_name=u"Autre type de circuit court")
    legumes_commercialisation_cl = models.CharField(max_length="100",
            null=True, blank=True,
            choices=CHOIX_COMMERCIALISATION_CIRCUIT_LONG,
            verbose_name=u"commercialisation en circuit long")
    legumes_cl_autres = models.CharField(
            max_length="255", null=True, blank=True,
            verbose_name=u"Autre type de circuit long")
    legumes_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    legumes_transformation = models.NullBooleanField(
            verbose_name=u"transformation")
    legumes_detail_transformation = models.TextField(null=True, blank=True,
            verbose_name=u"produits transformés")
    legumes_systeme_forcage_endives = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_SYSTEME_FORCAGE_ENDIVES,
            verbose_name=u"système de forçage d'endives")

    # Culture
    culture = models.NullBooleanField(verbose_name=u"grandes cultures")
    culture_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    culture_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    culture_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    culture_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    culture_transformation = models.NullBooleanField(
            verbose_name=u"transformation")
    culture_detail_transformation = models.TextField(null=True, blank=True,
            verbose_name=u"détail transformation")
    culture_stockage_ferme = models.NullBooleanField(
            verbose_name=u"stockage à la ferme")

    # Vergers basse tige
    vergers_bt = models.NullBooleanField(verbose_name=u"vergers basse tige")
    vergers_bt_fruits = models.CharField(max_length="255",
            null=True, blank=True,
            verbose_name=u"fruits vergers basse tige")
    vergers_bt_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    vergers_bt_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    vergers_bt_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    vergers_bt_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    vergers_bt_transformation = models.NullBooleanField(
            verbose_name=u"transformation")
    vergers_bt_detail_transformation = models.TextField(null=True, blank=True,
            verbose_name=u"détail transformation")

    # Vergers haute tige
    vergers_ht = models.NullBooleanField(verbose_name=u"vergers haute tige")
    vergers_ht_pommes = models.BooleanField(
            verbose_name=u"pommes")
    vergers_ht_poires = models.BooleanField(verbose_name=u"poires")
    vergers_ht_prunes = models.BooleanField(verbose_name=u"prunes")
    vergers_ht_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    vergers_ht_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    vergers_ht_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    vergers_ht_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    vergers_ht_transformation = models.NullBooleanField(
            verbose_name=u"transformation")
    vergers_ht_detail_transformation = models.TextField(null=True, blank=True,
            verbose_name=u"détail transformation")

    # Autres productions
    autres_petits_fruits = models.NullBooleanField(
            verbose_name=u"petits fruits")
    autres_petits_fruits_detail = models.TextField(null=True, blank=True,
            verbose_name=u"détail des petits fruits")
    autres_plants_maraichers = models.NullBooleanField(
            verbose_name=u"plants maraîchers")
    autres_plants_detail = models.TextField(null=True, blank=True,
            verbose_name=u"types de plants maraîchers")
    autres_nb_plants_an = models.CharField(null=True, blank=True,
            max_length="255", 
            verbose_name=u"nombre de plants produits par an")
    autres_ppam = models.NullBooleanField(
            verbose_name=u"PPAM")
    autres_ppam_detail = models.TextField(null=True, blank=True,
            verbose_name=u"détail PPAM")
    autres_prod_veg = models.NullBooleanField(
            verbose_name=u"autres productions végétales spécialisées")
    autres_prod_veg_detail = models.TextField(null=True, blank=True,
            verbose_name=u"détail de ces productions")
    autres_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    autres_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    autres_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    autres_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    autres_transformation = models.NullBooleanField(
            verbose_name=u"transformation")
    autres_detail_transformation = models.TextField(null=True, blank=True,
            verbose_name=u"détail transformation")
 
    # Lait
    lait_vaches = models.NullBooleanField(
            verbose_name=u"vaches laitières")
    lait_nb_vaches = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de vaches laitières")
    lait_vaches_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de vaches laitières en conversion")
    lait_quota_laiterie = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"quota laiterie (en L)")
    lait_quota_vente_directe = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"quota vente directe (en L)")
    lait_prod_laiterie = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"production lait bio laiterie (en L)")
    lait_prod_vente_directe = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"production lait bio vente directe (en L)")
    lait_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    lait_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    lait_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    lait_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    lait_laiterie = models.ForeignKey(ref.Laiterie, 
            null=True, blank=True,
            verbose_name=u"laiterie")
    lait_transformation_ferme = models.NullBooleanField(
            verbose_name=u"transformation de lait à la ferme")
    lait_transformation_presta = models.NullBooleanField(
            verbose_name=u"transformation de lait en prestation")
    lait_atelier_lait = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier lait")
    lait_produits_transformes = models.TextField(null=True, blank=True,
            verbose_name=u"produits laitiers transformés")
    # viande issue du troupeau laitier
    lait_reformes_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb réformes vendues en circuit long")
    lait_genisses_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb génisses vendues en circuit long")
    lait_taurillons_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb taurillons vendues en circuit long")
    lait_boeufs_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb boeufs vendus en circuit long")
    lait_caissette = models.NullBooleanField(
            verbose_name=u"vente caissette animaux laitiers")
    lait_adultes_cc = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb d'adultes laitiers vendus en circuit court / an")
    lait_veaux_cc = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb de veaux laitiers vendus en circuit court / an")
    lait_abattoirs_vl = models.ManyToManyField(ref.Abattoir,
            null=True, blank=True, related_name="abattoir_vl",
            verbose_name=u"Abattoirs des vaches laitières")
    lait_decoupe_presta = models.NullBooleanField(
            verbose_name=u"découpe en prestation")
    lait_atelier_decoupe = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier decoupe")

    # Lait de chevre
    lait_chevres = models.NullBooleanField(
            verbose_name=u"chèvres")
    lait_nb_chevres = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de chèvres")
    lait_chevres_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de chèvres en conversion")
    lait_chevres_prod = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"litres de lait de chèvre produits par an")
    lait_chevres_transformation_ferme = models.NullBooleanField(
            verbose_name=u"transformation de lait de chèvre à la ferme")
    lait_chevres_transformation_presta = models.NullBooleanField(
            verbose_name=u"transformation de lait de chèvre en prestation")
    lait_chevres_atelier_lait = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier lait de chèvre")
    lait_chevres_type_produits = models.TextField(null=True, blank=True,
            verbose_name=u"type produits laitiers")
    lait_chevres_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    lait_chevres_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    lait_chevres_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    lait_chevres_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
 
    # Viande bovine
    bovine = models.NullBooleanField(
            verbose_name=u"viande bovine")
    bovine_nb_va = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre VA")
    bovine_nb_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"VA conversion")
    bovine_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    bovine_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    bovine_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    bovine_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    bovine_reformes_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb réformes vendues en circuit long")
    bovine_genisses_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb génisses vendues en circuit long")
    bovine_taurillons_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb taurillons vendues en circuit long")
    bovine_boeufs_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb boeufs vendus en circuit long")
    bovine_caissette = models.NullBooleanField(
            verbose_name=u"vente caissette animaux race à viande")
    bovine_adultes_cc = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb d'adultes race à viande"
                         u" vendus en circuit court / an")
    bovine_veaux_cc = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb de veaux race à viande"
                         u" vendus en circuit court / an")
    bovine_abattoirs_va = models.ManyToManyField(ref.Abattoir,
            null=True, blank=True, related_name="abattoir_va",
            verbose_name=u"Abattoirs VA")
    bovine_decoupe_presta = models.NullBooleanField(
            verbose_name=u"découpe en prestation")
    bovine_atelier_va = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier va")

    # Porcs
    porcs = models.NullBooleanField(
            verbose_name=u"porcs bio")
    porcs_nb_truies = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de truies")
    porcs_nb_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"porcs conversion")
    porcs_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    porcs_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    porcs_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    porcs_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    porcs_transformation_ferme = models.NullBooleanField(
            verbose_name=u"transformation à la ferme")
    porcs_produits_transformes = models.TextField(null=True, blank=True,
            verbose_name=u"principaux produits transformés")
    porcs_abattoirs = models.ManyToManyField(ref.Abattoir,
            null=True, blank=True, related_name="abattoir_porcs",
            verbose_name=u"Abattoirs porcs")
    porcs_decoupe_presta = models.NullBooleanField(
            verbose_name=u"découpe en prestation")
    porcs_atelier = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier porcs")

    # Moutons
    moutons = models.NullBooleanField(
            verbose_name=u"moutons bio")
    moutons_nb_brebis = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nombre de brebis")
    moutons_nb_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"brebis conversion")
    moutons_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    moutons_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    moutons_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    moutons_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    moutons_transformation_ferme = models.NullBooleanField(
            verbose_name=u"transformation à la ferme")
    moutons_abattoirs = models.ManyToManyField(ref.Abattoir,
            null=True, blank=True, related_name="abattoir_moutons",
            verbose_name=u"Abattoirs agneaux")
    moutons_decoupe_presta = models.NullBooleanField(
            verbose_name=u"découpe en prestation")
    moutons_atelier = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier agneaux")

    # Petites animaux
    animaux_lapins = models.NullBooleanField(
            verbose_name=u"production de lapins")
    animaux_volailles = models.NullBooleanField(
            verbose_name=u"production de volailles")
    animaux_volailles_chair = models.TextField(null=True, blank=True,
            verbose_name=u"volailles de chair")
    animaux_nb_conversion = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"petits animaux conversion")
    animaux_nb_poulets = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb poulets (produits par an)")
    animaux_nb_lapins = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb lapins")
    animaux_nb_autres_volailles = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb autres volailles")
    animaux_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    animaux_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    animaux_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    animaux_hors_vd = models.TextField(null=True, blank=True,
            verbose_name=u"commercialisation hors vente directe")
    animaux_atelier = models.CharField(max_length="100",
            null=True, blank=True, choices=CHOIX_TYPE_ATELIER,
            verbose_name=u"type atelier petits animaux")
    animaux_decoupe_presta = models.NullBooleanField(
            verbose_name=u"découpe en prestation")
    animaux_lieu_abattage = models.TextField(null=True, blank=True,
            verbose_name=u"lieu d'abattage et transformation")

    # Autres viandes
    animaux_autres_prod = models.NullBooleanField(
            verbose_name=u"Autres productions animales bio")
    animaux_details_autres_prod = models.TextField(null=True, blank=True,
            verbose_name=u"Autres productions")

    # Œufs
    oeufs_production_bio = models.NullBooleanField(
            verbose_name=u"production d'œufs bio")
    oeufs_nb_pondeuses = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb de pondeuses")
    oeufs_nb_oeufs = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb d'œufs")
    oeufs_acheteur_cl = models.ForeignKey(ref.AcheteurOeufs,
            null=True, blank=True,
            verbose_name=u"acheteur circuit long")
    oeufs_comm_vd = models.BooleanField(
            verbose_name=u"commercialisation en vente directe")
    oeufs_comm_cc = models.BooleanField(
            verbose_name=u"en circuit court")
    oeufs_comm_cl = models.BooleanField(
            verbose_name=u"en circuit long")
    oeufs_circuits_courts = models.TextField(null=True, blank=True,
            verbose_name=u"circuits courts œufs")

    # Miel
    miel_production_bio = models.NullBooleanField(
            verbose_name=u"apiculture")
    miel_nb_ruches = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"nb de ruches")

    # Conventionnel
    conv_prod_animales = models.NullBooleanField(
            verbose_name=u"productions animales conventionnelles")
    conv_detail_prod_animales = models.TextField(null=True, blank=True,
            verbose_name=u"détail de ces productions animales")
    conv_prod_vegetales = models.NullBooleanField(
            verbose_name=u"productions végétales conventionnelles")
    conv_detail_prod_vegetales = models.TextField(null=True, blank=True,
            verbose_name=u"détail de ces productions végétales")
    conv_existe_prod_sans_valorisation = models.NullBooleanField(
            verbose_name=u"productions sans valorisation possible en bio")
    conv_prod_sans_valorisation = models.TextField(null=True, blank=True,
            verbose_name=u"productions sans valorisation possible en bio")

    # Formation
    formation_annee = models.NullBooleanField(
            verbose_name=u"avez-vous participé à une formation"
                         u" au cours de l'année écoulée ?")
    formation_annee_laquelle = models.ForeignKey(
            ref.Formation, null=True, blank=True,
            on_delete=models.SET_NULL,
            verbose_name=u"si oui, laquelle ?")
    formation_annee_commentaire = models.TextField(
            null=True, blank=True,
            verbose_name=u"commentaires")
    formation_avenir = models.NullBooleanField(
            verbose_name=u"avez-vous des besoins en formation à l'avenir ?")
    formation_avenir_type = models.ForeignKey(
            ref.TypeFormation, null=True, blank=True,
            on_delete=models.SET_NULL,
            verbose_name=u"si oui, lesquelles ?")
    formation_avenir_details = models.TextField(
            null=True, blank=True,
            verbose_name=u"détails")

    # Infos assolement
    sau_agri = models.IntegerField(null=True, blank=True,
                        verbose_name=u"SAU renseignée")
    total_surfaces_assolement = models.IntegerField(null=True, blank=True,
                 verbose_name=u"total des surfaces d'assolement"
                              u" saisies ci-dessous",
                 help_text=u"Note: ce total n'est mis à jour qu'après"
                           u" enregistrement de la page")

    def __unicode__(self):
        """
        Retourne une représentation unicode de la structure.
        """
        return u"Données de %s pour l'année %s" \
                    % (self.agriculteur_bio, self.annee)

    def agri_id(self):
        if self.agriculteur_bio:
            return self.agriculteur_bio.id
        return u""

    agri_id.short_description = u"ID agriculteur bio"
    agri_id.verbose_name = u"ID agriculteur bio"

    def get_absolute_url(self):
        """
        Retourne l'URL du formulaire de modification de l'objet.
        """
        return reverse(
                'admin:%s_%s_change' % (self._meta.app_label,
                                        self.__class__.__name__.lower()),
                args=[self.pk])

    def get_assolements(self):
        """
        Retourne une représentation de tous les assolements associés à l'objet
        sous un format adapté à l'export CSV (chaque représentation
        d'assolement est séparé par une nouvelle ligne)
        """
        repr_assolements = []
        for a in self.assolement_set.all():
            repr_assolements.append(a.repr_export())
        return u'\r'.join(repr_assolements)

    get_assolements.verbose_name = u"assolements"



class Assolement(models.Model):
    """
    Informations d'assolement liées à DonneesBio (observatoire Gabnor).
    """
    donneesbio = models.ForeignKey('DonneesBio',
                                   verbose_name=u"objet donneesbio")
    culture = models.ForeignKey(ref.Culture, verbose_name=u"culture")
    detail_culture = models.CharField(max_length="255", null=True, blank=True,
            verbose_name=u"détail des cultures (si nécessaire)",
            help_text=u"Complétez si nécessaire la description des cultures")
    surface_bio = models.DecimalField(null=True, blank=True,
                                      max_digits=10, decimal_places=4,
                                      verbose_name=u"surface bio")
    surface_c1 = models.DecimalField(null=True, blank=True,
                                     max_digits=10, decimal_places=4,
                                     verbose_name=u"surface C1")
    surface_c2 = models.DecimalField(null=True, blank=True,
                                     max_digits=10, decimal_places=4,
                                     verbose_name=u"surface C2")
    surface_c3 = models.DecimalField(null=True, blank=True,
                                     max_digits=10, decimal_places=4,
                                     verbose_name=u"surface C3")
    surface_conv = models.DecimalField(null=True, blank=True,
                                       max_digits=10, decimal_places=4,
                                       verbose_name=u"surface conventionnelle")
    tonnage = models.FloatField(null=True, blank=True,
                                verbose_name=u"tonnage")
    client = models.ForeignKey(ref.ClientCulture, null=True, blank=True,
                               verbose_name=u"client",
                               on_delete=models.SET_NULL)
    prix = models.FloatField(null=True, blank=True,
                             verbose_name=u"Prix")
    type_circuit = models.CharField(max_length="10", null=True, blank=True,
                                    choices=CHOIX_TYPE_CIRCUIT)

    def __unicode__(self):
        """
        """
        return u"Assolement de %s" % self.culture

    def repr_export(self):
        """
        Retourne une représentation de l'objet adaptée à l'export CSV
        (permet de faire figurer les assolements dans l'export d'un objet
        donneesbio)
        Note: le séparateur du CSV peut être utilisé car la valeur du champ
        sera entourée de guillements « " ».
        """
        return u"culture: %s; détail: %s; bio: %s; c1: %s; c2: %s; c3: %s;"\
               u" conv: %s; tonnage: %s; client: %s; prix: %s;"\
               u" type de circuit: %s" \
                % (self.culture, self.detail_culture or '',
                   self.surface_bio or '', self.surface_c1 or '',
                   self.surface_c2 or '', self.surface_c3 or '',
                   self.surface_conv or '',
                   self.tonnage or '', self.client or '',
                   self.prix or '', self.get_type_circuit_display() or '')



class Personne(Contact, models.Model):
    """
    Personne physique dépendant d'une structure.
    .
    Attributs et méthodes hérités de la classe abstraite Contact (hérités aussi
    par Structure) pour permettre le partage du même panier avec les modèles
    basés sur Structures.
     - basket_name
     - classe_form_actions_panier
     - attrs_publipostage
     - adresse_electronique()
     - get_courriels()
     - liste_contacts_associes()
    """
    class Meta:
        ordering = ['nom','prenom']

    # Champs à inclure dans un export CSV complet
    attrs_export = [ 'civilite', 'nom', 'prenom', 'structure', 'adresse',
            'commune', 'code_postal', 'courrier_structure',
            'tel_fixe', 'tel_mobile', 'fax', 'courriel',
            'pas_de_courriel', 'date_creation', 'date_derniere_modification',
            'reseaux_engagements', 'date_creation',
            'date_derniere_modification', 'utilisateur_derniere_modification' ]

    civilite = models.CharField(max_length="10", null=True, blank=True,
                                choices=CHOIX_CIVILITE,
                                verbose_name=u"civilité")
    nom = models.CharField(max_length=255)
    prenom = models.CharField(max_length=255, verbose_name=u"prénom")
    structure = models.ForeignKey("Structure", verbose_name=u"affilié à")
    fonction = models.CharField(max_length="255", null=True, blank=True,
                                verbose_name=u"fonction")
    adresse = models.TextField(blank=True, verbose_name=u"adresse personnelle",
       help_text="Saisissez l'adresse de la personne, sur 4 lignes maximum.")
    commune = models.ForeignKey(ref.Commune, null=True, blank=True)
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(max_length=10, null=True, blank=True,
           help_text=u"Laisser vide pour utiliser celui associé à la commune")
    # la ville contient également le code cedex (à utiliser pour l'adresse)
    ville = models.CharField(max_length="255", null=True, blank=True,
               help_text=u"Nom de ville avec code cédex si nécessaire, "\
                         u" si différent du nom de la commune.")
    courrier_structure = models.BooleanField(default=False,
            verbose_name=u"Recevoir les courriers addressés à la structure",
            help_text=u"Cochez cette case si la personne doit recevoir"
                      u" les courriers papier envoyés à la structure.")
    tel_fixe = models.CharField(max_length=100, blank=True,
                                verbose_name=u"téléphone fixe")
    tel_mobile = models.CharField(max_length=100, blank=True,
                                  verbose_name=u"téléphone mobile")
    fax = models.CharField(max_length=100, null=True, blank=True)
    courriel = models.CharField(max_length=255, blank=True)
    pas_de_courriel = models.BooleanField(default=False,
            verbose_name=u"pas d'envoi automatique de courriel",
            help_text=u"Cochez cette case si la personne ne souhaite"
                      u" pas recevoir de courriel envoyé automatiquement.")
    reseaux_engagements = models.TextField(null=True, blank=True,
            verbose_name=u"Appartenance à des réseaux et engagements")

    date_creation = models.DateField(editable=False,
            verbose_name=u"date de saisie de l'enregistrement")
    date_derniere_modification = models.DateField(editable=False,
            verbose_name=u"date de dernière modification de informations")
    utilisateur_derniere_modification = models.ForeignKey(auth_models.User,
            editable=False, verbose_name=u"dernière modification par")
    
    def __unicode__(self):
        """
        Retourne une représentation unicode de la structure.
        """
        if not self.nom and not self.prenom:
            return u"Personne non nommée #%s" % self.id
        return u"%s %s" % (self.prenom, self.nom)
    
    def get_absolute_url(self):
        """
        Retourne l'URl associée à l'objet.
        """
        return reverse('admin:%s_%s_change' \
                % (self._meta.app_label, self.__class__.__name__.lower()),
                args =[self.pk])
    
    def nom_prenom(self):
        """
        Retourne les nom et prénom de la personne sous forme de chaîne de
        caractères.
        """
        return u"%s %s" % (self.nom.upper(), self.prenom)
    
    def contenu_etiquette(self):
        """
        Retourne le contenu à afficher dans l'étiquette sous forme de liste
        dont chaque élément représente une ligne.
        """
        contenu = []
        # On doit remplacer & par &amp; à cause d'un problème dans reportlab
        # qui génère une erreur XML lorsque le nom contient un "&"
        contenu.append(self.nom_prenom().replace("&", "&amp;"))
        contenu.extend(self.adresse.split('\n'))
        contenu.append("%s %s" % (self.code_postal or '', self.ville or ''))
        return contenu
    
    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresses email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        if self.nom:
            nom = u"%s %s" % (self.prenom, self.nom)
        else:
            nom = u""
        return (nom, self.courriel)

    def get_courriels(self):
        """
        Méthode commune aux structures et personnes: retourne la liste des
        adresses emails (dans Personne il n'y en a qu'une)
        """
        adr = self.adresse_electronique()
        if adr and not self.pas_de_courriel:
            return [ adr ]
        return []

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse,
        mieux vaut le gérer manuellement).
        """
        if self.id is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        super(Personne, self).save(*args, **kwargs)




### --------------------------------------------------------------------------
# Classes secondaires
#

class RendezVous(models.Model):
    """
    Rendez-vous de suivi avec un agriculteur.
    """
    class Meta:
        ordering = ['date']
        get_latest_by = 'date'
    
    date = models.DateField(db_index=True)
    technicien = models.ForeignKey(auth_models.User,
                    null=True, blank=True, on_delete=models.SET_NULL)
    commentaire = models.TextField(null=True, blank=True)

    def __unicode__(self):
        """
        Retourne une représentation unicode de l'objet.
        (méthode utilisée pour affiche l'objet RendezVous de base
        lors de la suppression d'un agriculteur et de ses dépendances, par
        exemple)
        """
        return u"Rendez-vous du %s" % self.date.strftime(LOCAL_DATE_FORMAT)


class RendezVousProjetBio(RendezVous):
    """
    Rendez-vous de suivi dans le cadre d'un projet bio.
    """
    class Meta:
        # ordering hérité
        verbose_name = "rendez-vous de suivi de projet bio"
        verbose_name_plural = "rendez-vous de projet bio"

#    attrs_export = [
#        'agriculteur',
#        ]

    attrs_export = [
        # identifiant projet
        'proj_id', 'agriculteur' , 'agri_id', 'techproj', 'description',
        # identifiant rendez vous
        'date', 'technicien', 'type', 'commentaire',
        # info projet
        'freins_identifies', 'suites_a_donner', 'remarques', 'etude_demandee',
        'date_etude', 'date_probable_certification', 'date_certification',
        'engagement_oc', 'engagement_oc_enjeu_eau', 'installation_bio', 'date_engagement_oc', 'commentaireproj',
        # info agriculteur
        'agri_cp', 'agri_commune', 'agri_adresse',
        # info exploitation
        'exp_type', 'exp_qual', 'exp_sau', 'exp_sau_bio', 'exp_sau_eau','exp_sau_eau_bio',
        'exp_quota_laitier', 'exp_laiterie', 'exp_nb_vache' ]


    basket_name =u"rendezvousbio"

    # Form fournissant les actions possibles sur le panier
    classe_form_actions_panier = contacts_forms.FormActionPanierDonnee

    type = models.CharField(max_length=250,
                            choices=contacts_forms.CHOIX_TYPES_RDV,
                            verbose_name=u"type de rendez-vous")
    projet_bio = models.ForeignKey('ProjetBio')

    # Hack pas très propre pour pouvoir afficher les rendez-vous sur une page
    # d'admin des modèles Agriculteur (puisqu'on ne pas l'afficher un inline
    # dans le inline du projet bio concerné)
    # On utilise on delete do nothing puisque la cascade devrait se faire déjà
    # via projet_bio.
    agriculteur = models.ForeignKey('Agriculteur', on_delete=models.DO_NOTHING)

    def save(self, *args, **kwargs):
        """
        """
        # on renseigne agriculteur automatiquement si ce n'est pas fait
        # pour permettre une utilisation normale sans s'occuper du hack
        # que constitue self.agriculteur
        try:
            self.agriculteur
        except Agriculteur.DoesNotExist:
            self.agriculteur = self.projet_bio.agriculteur
        super(RendezVousProjetBio, self).save(*args, **kwargs)

    def techproj(self):
        return self.projet_bio.technicien
    techproj.verbose_name=u"Suivi par"
    def description(self):
        return self.projet_bio.description
    def freins_identifies(self):
        return self.projet_bio.freins_identifies
    def suites_a_donner(self):
        return self.projet_bio.suites_a_donner
    def remarques(self):
        return self.projet_bio.remarques
    def etude_demandee(self):
        return self.projet_bio.etude_demandee
    def date_etude(self):
        return self.projet_bio.date_etude
    def date_probable_certification(self):
        return self.projet_bio.date_probable_certification
    def date_certification(self):
        return self.projet_bio.date_certification
    def engagement_oc(self):
        return self.projet_bio.engagement_oc
    def engagement_oc_enjeu_eau(self):
        return self.projet_bio.engagement_oc_enjeu_eau
    def installation_bio(self):
        return self.projet_bio.installation_bio
    def date_engagement_oc(self):
        return self.projet_bio.date_engagement_oc
    def commentaireproj(self):
        return self.projet_bio.commentaire
    commentaireproj.verbose_name=u"Commentaire"

#        'agri_nom', 'agri_cp', 'agri_ville', 'agri_adresse',
    def agri_cp(self):
        return self.projet_bio.agriculteur.code_postal
    agri_cp.verbose_name=u"Code postal"
    def agri_commune(self):
        return self.projet_bio.agriculteur.commune
    agri_commune.verbose_name=u"Commune"
    def agri_adresse(self):
        return self.projet_bio.agriculteur.adresse
    agri_adresse.verbose_name=u"Adresse"

    def exp_type(self):
        return self.projet_bio.agriculteur.type_activite
    exp_type.verbose_name=u"Type d'exploitation"

    def exp_qual(self):
        return ','.join([ u"%s" % qual for qual in self.projet_bio.agriculteur.qualificatifs_complementaires.all() ])

    exp_qual.verbose_name=u"Qualificatifs complémentaires"
    def exp_sau(self):
        return self.projet_bio.agriculteur.sau
    exp_sau.verbose_name=u"S.A.U."
    def exp_sau_bio(self):
        return self.projet_bio.agriculteur.sau_bio
    exp_sau_bio.verbose_name=u"S.A.U. Bio"
    def exp_sau_eau(self):
        return self.projet_bio.agriculteur.sau_eau
    exp_sau_eau.verbose_name=u"S.A.U. Enj. eau"
    def exp_sau_eau_bio(self):
        return self.projet_bio.agriculteur.sau_eau_bio
    exp_sau_eau.verbose_name=u"S.A.U. Bio Enj. eau"
    def exp_quota_laitier(self):
        return self.projet_bio.agriculteur.quota_laitier
    exp_quota_laitier.verbose_name=u"Quota laitier"
    def exp_laiterie(self):
        return self.projet_bio.agriculteur.laiterie
    exp_laiterie.verbose_name=u"Laiterie"
    def exp_nb_vache(self):
        return self.projet_bio.agriculteur.nb_vaches_allaitantes
    exp_nb_vache.verbose_name=u"Nb vaches all."


    def agri_id(self):
        if self.agriculteur:
            return self.agriculteur.id
        return u""
    agri_id.short_description = u"ID agriculteur"
    agri_id.verbose_name = u"ID agriculteur"

    def proj_id(self):
        if self.projet_bio:
            return self.projet_bio.id
        return u""
    proj_id.short_description = u"ID Projet"
    proj_id.verbose_name = u"ID Projet"

    def __unicode__(self):
        """
        Retourne une représentation unicode de l'objet.
        """
        return u"Rendez-vous du %s pour %s" \
                % (self.date.strftime(LOCAL_DATE_FORMAT), self.projet_bio)


class RendezVousSuivi(RendezVous):
    """
    Rendez-vous de suivi avec un agriculteur Bio.
    """
    class Meta:
        # ordering hérité
        verbose_name = "rendez-vous de visite annuelle"
        verbose_name_plural = "rendez-vous de visite annuelle"

    agriculteur_bio = models.ForeignKey('AgriculteurBio',
                                        verbose_name=u"agriculteur bio")

    def __unicode__(self):
        """
        Retourne une représentation unicode de l'objet.
        """
        return u"Rendez-vous de visite annuelle du %s" % \
                                self.date.strftime(LOCAL_DATE_FORMAT)


class AdhesionGabnor(models.Model):
    """
    Adhésion Gabnor.
    """
    class Meta:
        ordering = ['date']
        get_latest_by = 'date'
        verbose_name = "adhésion GABNOR"
        verbose_name_plural = "adhésions GABNOR"

    # Champs à inclure dans un export CSV complet
    attrs_export = [ 'agriculteur_bio', 'date', 'montant', ]

    # Le panier où conserver les objets sélectionnés
    basket_name = u"adhésions gabnor"

    # Form fournissant les actions possibles sur le panier
    classe_form_actions_panier = contacts_forms.FormActionPanierDonnee
    
    agriculteur_bio = models.ForeignKey('AgriculteurBio')
    date = models.DateField(db_index=True)
    montant = models.DecimalField(max_digits=8, decimal_places=2,
                      null=True, blank=True,
                      verbose_name=u"montant de l'adhésion (en euros)")
    
    def __unicode__(self):
        """
        Représentation unicode de l'objet (utilisé notamment pour affichage
        dans l'objet AgriculteurBio).
        """
        return u"Adhésion du %s de %.2f EUR de %s" \
                % (self.date.strftime(LOCAL_DATE_FORMAT).decode('utf8'),
                   self.montant, self.agriculteur_bio)
    
    def save(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        """
        super(AdhesionGabnor, self).save(*args, **kwargs)
        self._update_dates_agriculteurbio()

    def delete(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        (cette méthode n'est pas appelée pour les suppressions en masse,
        mais ça n'arrivera pas dans un usage normal de l'application).
        """
        super(AdhesionGabnor, self).delete(*args, **kwargs)
        self._update_dates_agriculteurbio()

    def _update_dates_agriculteurbio(self):
        """
        Calcule et sauvegarde les dates de première et dernière adhésion,
        repectivement date_adhesion et date_renouvellement dans l'objet
        AgriculteurBio.
        """
        adhesions_qs = AdhesionGabnor.objects\
                            .filter(agriculteur_bio=self.agriculteur_bio)\
                            .order_by('date')
        if len(adhesions_qs) != 0:
            premiere_adhesion = adhesions_qs[0]
        else:
            premiere_adhesion = None
        if len(adhesions_qs) > 1:
            # le renouvellement est la dernière adhésion en date
            renouvellemenent = adhesions_qs[len(adhesions_qs)-1]
        else:
            renouvellemenent = None
        self.agriculteur_bio.premiere_adhesion = premiere_adhesion
        self.agriculteur_bio.renouvellement_adhesion = renouvellemenent
        self.agriculteur_bio.save()


#
### --------------------------------------------------------------------------
