# -*- coding: utf-8 -*-
#
# Ce fichier regroupe les classes et fonctions annexes utilisées dans les vues
# de l'application.
#
##

import datetime, os, locale, types, codecs

import reportlab
import reportlab.platypus as rlp
import reportlab.lib.styles as rlstyles
import reportlab.lib.enums as rlenums
from reportlab.lib.units import cm

from basecontacts_gabnor import logger
import basecontacts_gabnor.contacts.models as contacts_models

from django.conf import settings
from django.db import models
import django.db

PAGE_HEIGHT=reportlab.rl_config.defaultPageSize[1]
PAGE_WIDTH=reportlab.rl_config.defaultPageSize[0]



### ----------------------------------------------------------------------------
# Transformation d'un agriculteur en agriculteur Bio 
#

def transforme_agriculteur(agriculteur):
    """
    Transforme un Agriculteur en AgriculteurBiologique
    avec des requêtes SQL de bas niveau (seule solution trouvée).
    .
    Utilisé à la fois dans l'action transformer_en_bio et la vue
    TransformeAgriculteur.
    """
    id_agri = agriculteur.agriculteur.id
    try:
        #contacts_models.AgriculteurBio.objects.get(pk=id_agri)
        agriculteur.agriculteurbio
    except contacts_models.AgriculteurBio.DoesNotExist:
        pass
    else: # est déjà un agriculteur bio
        return 0
    # Création de l'agriculteur bio
    cursor = django.db.connection.cursor()
    cursor.execute(
           "INSERT INTO %s_agriculteurbio (agriculteur_ptr_id) VALUES (%s)" \
           % (agriculteur._meta.app_label, id_agri))
    django.db.transaction.commit_unless_managed()
    return 1

def transforme_structure(structure):
    """
    Transforme un Structure en Agriculteur
    avec des requêtes SQL de bas niveau (seule solution trouvée).
    """
    id_struct = structure.id
    try:
        #contacts_models.StructureBio.objects.get(pk=id_struct)
        structure.agriculteur
    except contacts_models.Agriculteur.DoesNotExist:
        pass
    else: # est déjà un agriculteur
        return 0
    # Création de l'agriculteur
    cursor = django.db.connection.cursor()
    cursor.execute(
        # le champs visibilite a une contrainte "not null", on rempli avec la valeur par defaut.
           "INSERT INTO %s_agriculteur (structure_ptr_id,visibilite) VALUES (%s,'personne')" \
           % (structure._meta.app_label, id_struct))
    django.db.transaction.commit_unless_managed()
    return 1



### ---------------------------------------------------------------------------
# Étiquettes en PDF
#

class EtiquettesPDF:
    """
    Classe permettant la génération simple de pages d'étiquettes d'adresses
    pour les modèles fournissant la méthode contenu_etiquette() retournant
    la liste des lignes constituant l'étiquette.
    """
    # Marges de la page
    margin_top = 0*cm
    margin_left = 0*cm
    margin_right = 0*cm
    margin_bottom = 0*cm

    # Dimensions des étiquettes (représentées par des frames)
    frame_width = 7.00*cm
    frame_height = 3.71*cm
    # Nombre de colonnes et lignes de frames par page
    frames_cols = int((PAGE_WIDTH-margin_left-margin_right)/frame_width)
    frames_rows = int((PAGE_HEIGHT-margin_top-margin_bottom)/frame_height)
    # Nombre de frames (donc d'étiquettes) par page
    fpp = frames_cols*frames_rows

    # Bricolage sur la hauteur des étiquettes pour régler un problème
    # sur l'étiquette du bas qui est trop haute sur les feuilles d'étiquettes
    # du Gabnor. (Fait ici pour ne pas impacter fpp)
	# On rajoute cette hauteur à chaque étiquette.
    frame_height += .04*cm

    # Le padding des frames, dans l'ordre demandé par rlp.Frame(),
    # c'est à dire left, bottom, right, top
    frame_padding = (10, 5, 5, 5)
    # Dessiner les bords des frames ?
    show_frame_boundary = 0

    # Première ligne des frames
    top_y = PAGE_HEIGHT-margin_top

    titre = u"Étiquettes"
    auteur = u""

    def __init__(self, file):
        """
        Calcule quelques informations avant sur le document à générer.
        """
        # Le fichier (ou l'objet émulant un fichier) où écrire
        self.file = file
        # Style des étiquettes
        self.style = rlstyles.getSampleStyleSheet()["Normal"]
        self.style.fontName = "Helvetica"
        self.style.fontSize = 11

    def build(self, objets):
        """
        Génère le document et l'écrit dans self.file.
        """
        # Création du Canvas
        self.canv = reportlab.pdfgen.canvas.Canvas(self.file)
        self.canv.setTitle(self.titre)
        self.canv.setAuthor(self.auteur)
        # Get on with it !
        page_frames = None
        for i in range(0, len(objets)):
            if i % self.fpp == 0: # On entame une nouvelle page
                if i != 0: # On génère la page qui vient d'être remplie
                    self.canv.showPage()
                # On crée la liste des frames de la nouvelle page
                page_frames = self._mk_page_frames(i//self.fpp)
            # On écrit l'étiquette dans la frame en cours
            page_frames[i % self.fpp].addFromList(
                    self._contenu_etiquette(objets[i]), self.canv)
        # Génération de la dernière page
        self.canv.showPage()
        # Génération du document
        self.canv.save()

    def _mk_page_frames(self, page_id):
        """
        Retourne une liste de frames pour une page.
        .
        page_id est le numéro de la page permettant d'assigner une identifiant
        unique à chaque frame, composé du numéro de page et du numéro de la
        frame dans cette page (même si ça n'est pas vraiment utilisé ensuite!).
        """
        frames = []
        for i in range(0, self.fpp):
            # remarque: *self.frame_padding ne fonctionne pas sous python2.5
            f = rlp.Frame(
                      self.margin_left+(i%self.frames_cols*self.frame_width),
                      self.top_y-(i/self.frames_cols+1)*self.frame_height,
                      self.frame_width,
                      self.frame_height,
                      self.frame_padding[0],
                      self.frame_padding[1],
                      self.frame_padding[2],
                      self.frame_padding[3],
                      id="%s%s" % (page_id, i),
                      showBoundary=self.show_frame_boundary
                      )
            frames.append(f)
        return frames

    def _contenu_etiquette(self, objet):
        """
        Retourne sous forme de listes de paragraphes le texte d'une étiquette
        qui sera écrit dans une frame.
        """
        return [ rlp.Paragraph(ligne, self.style)
                 for ligne in objet.contenu_etiquette() if ligne ]



### ---------------------------------------------------------------------------
# Base pour l'export CSV
#

class ExportCSV:
    """
    Classe abstraite de base permettant la génération de fichier CSV.
    .
    Les classes filles doivent préciser dans « attr_liste » le nom de
    l'attribut de classe du modèle donnant la liste des attributs du modèle
    devant figurer dans le CSV.
    Ces attributs peuvent être un attribut classique ou une méthode (on
    exportera alors le résultat de l'appel à cette méthode). Nota Bene: La
    méthode doit posséder un argument verbose_name.
    .
    La génération se fait sur un fichier (qui peut être un StringIO).
    """
    # Le caractère de séparation à utiliser dans le fichier généré
    SEP = u";"
    # L'attribut du modèle fournissant la liste des attributs à exporter.
    # Il DOIT être précisé dans les classes filles.
    attr_liste = None

    def __init__(self, file):
        """
        Constructeur: ne fait que conserver le fichier de sortie.
        """
        self.out = file

    def _quote_csv(self, value):
        """
        Retourne la chaîne de caractère value en ajoutant les « " » si
        nécessaire.
        """
        if value.find('"'):
            value = value.replace('"','""')
        if value.find('\r\n'):
            # Excel termine la ligne lorsqu'une nouvelle ligne est présente
            # dans un champ, même entre guillements. Tentative de contournement
            # (non testée) en remplaçant \r\n par \n 
            # voir http://stackoverflow.com/questions/1241220/generating-csv-file-for-excel-how-to-have-a-newline-inside-a-value
            value = value.replace('\r\n','\r')
        return '"%s"' % value

    def _get_repr(self, objet, attribute_name):
        """
        Retourne la représentation de l'attribut « objet » à inclure dans le
        fichier d'export, selon le type d'objet.
        Retourne une chaîne vide si l'objet ne possède pas cet attribut (pour
        pouvoir exporter des Structures, Agriculteur et AgriculteurBio en même
        temps).
        """
        value = None
        # Lecture de la valeur
        try:
            attribute_value = getattr(objet, attribute_name)
        except AttributeError: 
            return u'' # plus efficace que d'ajouter un default à getattr
        # Extraction en un type simple
        if isinstance(attribute_value, bool):
            # On convertit les booleéns en VRAI/FAUX pour les tableurs fr
            value = attribute_value and 'VRAI' or 'FAUX'
        elif isinstance(attribute_value, types.MethodType):
            # S'il s'agit d'une méthode, on utilise ce qu'elle retourne
            value = attribute_value()
        elif hasattr(objet, '_meta')\
                and objet._meta.get_field_by_name(attribute_name)[3] is True:
            # (get_field_by_name est non documentée. Voir
            # help(django.db.models.options.Options) dans un shell).
            # En cas de ManyToManyField, on retourne la liste des
            # représentations des objets, séparées par une virgule.
            value = ','.join([ unicode(o) for o in attribute_value.all() ])
        else:
            value = attribute_value
        # Conversion si nécessaire
        if value is None:
            # On substitue une chaîne vide à None
           value = u'' 
        elif isinstance(value, (datetime.date, datetime.datetime)):
            # Si value est une date, on la convertit dans notre format préféré
            value = value.strftime(settings.FORMAT_DATE_EXPORT)
        # On retourne la représentation unicode qu'on va quoter pour le CSV
        return self._quote_csv(unicode(value))

    def _ecrit_champs(self, objet):
        """
        Écrit les champs de l'objet donné dans le fichier de sortie.
        .
        L'objet doit posséder un attribut (désigné par notre attribut de classe
        attr_liste) qui donne la liste des champs à inclure dans l'export.
        """
        ligne = self.SEP.join([ self._get_repr(objet, champ)
                                for champ in self.attrs_export ])
        self.out.write((ligne+u"\r\n").encode('utf8'))

    def _get_verbose_names(self, obj, attrs):
        """
        Retourne la liste des verbose_name des champs sélectionnés.
        """
        titres = []
        for a in attrs:
            try:
                field = obj._meta.get_field(a)
            except models.fields.FieldDoesNotExist:
                # On accepte aussi les méthodes, qui n'apparaissent pas
                # dans Options.get_field.
                field = getattr(obj, a)
            try:
                vname = field.verbose_name
            except AttributeError:
                # Si l'attribut, ou surtout la méthode, n'a pas de
                # verbose_name, on en construit un à partir de son nom.
                vname = a.replace('_', ' ')
            titres.append(vname.capitalize())
        return self.SEP.join(titres)

    def generer(self, object_list):
        """
        Génère le fichier CSV à partir d'une liste d'objets.
        .
        Note: pour les structures, la liste des attributs à exporter est
        systématiquement celle du modèle AgriculteurBio et les champs non
        trouvés seront exportés vides (ceci permet d'exporter en même temps tous
        les objets Structure, Agriculteur, ou AgriculteurBio).
        """
        if len(object_list) == 0:
            return
        # On se base sur le premier objet pour récupérer le nom verbeux des
        # attributs.
        obj = object_list[0]
        # Si l'objet désigne une autre classe dans attrs_verbose_from, alors on
        # utilise un objet de cette classe pour récupérer les noms verbeux.
        nom_classe_verb = getattr(obj, 'attrs_verbose_from', None)
        if nom_classe_verb:
            classe_verb = contacts_models.__dict__[nom_classe_verb]
            verb_obj = classe_verb()
        else:
            verb_obj = obj
        try:
            self.attrs_export = getattr(obj, self.attr_liste)
        except AttributeError:
            raise contacts_models.AppImproperlyConfigured(
                    (u"Le modèle %s ne possède pas l'attribut %s"
                     u" fournissant la liste des champs à exporter") \
                             % (type(obj), self.attr_liste))
        titres = self._get_verbose_names(verb_obj, self.attrs_export)
        # Commencer le fichier par le BOM UTF8 est sensé aider Excel à 
        # comprendre qu'il s'agit d'UTF8
        self.out.write(codecs.BOM_UTF8)
        self.out.write((titres+u"\r\n").encode('utf8'))
        for obj in object_list:
            self._ecrit_champs(obj)



### ---------------------------------------------------------------------------
# Export CSV complet
#

class ExportCSVComplet(ExportCSV):
    """
    Classe effectuant un export CSV des attributs listés dans l'attribut de
    classe « attrs_export » du modèle.
    """
    # L'attribut du modèle fournissant la liste des attributs à exporter.
    attr_liste = 'attrs_export'



### ---------------------------------------------------------------------------
# Export CSV pour publipostage
#

class ExportCSVPublipostage(ExportCSV):
    """
    Classe effectuant un export CSV des attributs listés dans l'attribut de
    classe « attrs_publipostage » du modèle.
    """
    # L'attribut du modèle fournissant la liste des attributs à exporter.
    attr_liste = 'attrs_publipostage'



### ---------------------------------------------------------------------------
